#
# Copyright 2010-2012 Universidad Complutense de Madrid
# 
# This file is part of PyEmir
# 
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

'''Recipe for the reduction of gain calibration frames.'''

import logging
import os.path

import numpy
import scipy.stats

import numina.qa
from numina.recipes import RecipeBase, provides, Parameter
#from emir.dataproducts import create_result

from emir.instrument.detector import CHANNELS, QUADRANTS

from ...dataproducts import create_result

from ...dataproducts import MasterGainMap, MasterRONMap

_logger = logging.getLogger("emir.recipes")


def multimap(fun, ll):
    if hasattr(ll, "__iter__"):
        return [multimap(fun, i) for i in ll]
    return fun(ll)

@provides(MasterGainMap, MasterRONMap)
class Recipe1(RecipeBase):
    '''Detector Gain Recipe.
    
    Recipe to calibrate the detector gain.
    '''
    __requires__ = [
#        Schema('resets', ProxyPath('/observing_block/result/resets'), 
#               'A list of paths to reset images'),
#        Schema('ramps', ProxyPath('/observing_block/result/ramps'), 
#               'A list of ramps'),
        Parameter('region', 'channel', 'Region used to compute: (full|quadrant|channel)')
    ]

    def __init__(self):
        super(Recipe1, self).__init__(
            author="Sergio Pascual <sergiopr@fis.ucm.es>",
            version="0.1.0"
        )

    def run(self, obresult):
        return {'products': [MasterGainMap(), MasterRONMap()]}
  
    
    def region_full(self):
        return [(slice(0, 2048), slice(0, 2048))]
    
    def region_quadrant(self):
        return QUADRANTS
    
    def region_channel(self):
        return CHANNELS

    def region(self):
        fun = getattr(self, 'region_%s' %  self.parameters['region'])  
        return fun()
    
    def setup(self):
        # Sanity check, check: all images belong to the same detector mode
        
        self.parameters['resets'] = map(lambda x: DiskImage(os.path.abspath(x)), 
                                             self.parameters['resets'])
        self.parameters['ramps'] = multimap(lambda x: DiskImage(os.path.abspath(x)), 
                                             self.parameters['ramps'])

    def old_run(self):
        channels = self.region()
        ramps = self.parameters['ramps']
        result_gain = numpy.zeros((len(ramps), len(channels)))
        result_ron = numpy.zeros_like(result_gain)
        
        for ir, ramp in enumerate(ramps):
            counts = numpy.zeros((len(ramp), len(channels)))
            variance = numpy.zeros_like(counts)
            for i, di in enumerate(ramp):
                di.open(mode='readonly')
                try:
                    for j, channel in enumerate(channels):    
                        counts[i][j] = di.data[channel].mean()
                        variance[i][j] = di.data[channel].var(ddof=1)
                finally:
                    di.close()

            for j, _ in enumerate(channels):
                ig, ron,_,_,_ = scipy.stats.linregress(counts[:,j], variance[:,j])

                result_gain[ir][j] = 1.0 / ig
                result_ron[ir][j] = ron
        _logger.info('running')
        gch_mean = result_gain.mean(axis=0)
        gch_var = result_gain.var(axis=0, ddof=1)
        rch_mean = result_ron.mean(axis=0)
        rch_var = result_ron.var(axis=0, ddof=1)
        
        cube = numpy.zeros((2, 2048, 2048))
         
        for gain, var, channel in zip(gch_mean, gch_var, channels):
            cube[0][channel] = gain
            cube[1][channel] = var
        
        result = create_result(cube[0], variance=cube[1])                                    
        
        return {'qa': numina.qa.UNKNOWN, 'gain': {'mean': list(gch_mean.flat), 
                                                  'var': list(gch_var.flat),
                                                  'image': result
                                                  },
                                          'ron': {'mean': list(rch_mean.flat), 
                                                  'var': list(rch_var.flat)},
                                                  }



#
# Copyright 2011-2012 Universidad Complutense de Madrid
# 
# This file is part of PyEmir
# 
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

'''
Beam switched-nodded image mode recipe of EMIR

'''

from numina.recipes import RecipeBase, Parameter, DataProductParameter
from numina.recipes.requirements import Requirement
from numina.recipes import provides, DataFrame

from emir.dataproducts import MasterBias, MasterDark, MasterBadPixelMask
from emir.dataproducts import MasterIntensityFlat
from emir.dataproducts import NonLinearityCalibration
from emir.dataproducts import SourcesCatalog

from .shared import DirectImageCommon
    
@provides(DataFrame, SourcesCatalog)
class NBImageRecipe(RecipeBase, DirectImageCommon):
    '''
    The effect of recording a series of stare images, with the same
    acquisition parameters, and taken by pointing the TS in cycles
    between two, or more, sky positions. Displacements are larger
    than the EMIR FOV, so the images have no common area. Used
    for sky subtraction


    **Observing modes:**

        * Nodded/Beamswitched images
    
    '''

    __requires__ = [
        DataProductParameter('master_bpm', MasterBadPixelMask, 
                  'Master bad pixel mask'),       
        DataProductParameter('master_bias', MasterBias, 'Master bias image'),
        DataProductParameter('master_dark', MasterDark, 'Master dark image'),
        DataProductParameter('nonlinearity', NonLinearityCalibration([1.0, 0.0]), 
                  'Polynomial for non-linearity correction'),
        DataProductParameter('master_intensity_ff', MasterIntensityFlat, 
                  'Master intensity flatfield'),
        Parameter('extinction', 0.0, 'Mean atmospheric extinction'),
        Requirement('instrument.detector.channels', 'List of channels'),
        Requirement('instrument.detector.shape', 'Detector shape'),
        # FIXME: this parameter is optional 
        Parameter('sources', None, 'List of x, y coordinates to measure FWHM',
                  optional=True),
        Parameter('offsets', None, 'List of pairs of offsets',
                  optional=True)
    ]

    def __init__(self):
        super(NBImageRecipe, self).__init__(
            author="Sergio Pascual <sergiopr@fis.ucm.es>",
            version="0.1.0"
        )

    def run(self, obresult):
        
        baseshape = self.parameters['instrument.detector.shape']
        amplifiers = self.parameters['instrument.detector.channels']
        offsets = self.parameters['offsets']

        return self.process(obresult, baseshape, amplifiers, 
                            offsets=offsets, target_is_sky=False,
                            stop_after=DirectImageCommon.FULLRED)





# Copyright 2008-2012 Universidad Complutense de Madrid
# 
# This file is part of PyEmir
# 
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
# 

''' Recipes for engineering modes.

'''

import logging

from numina.recipes import RecipeBase, Parameter, provides, requires
from numina.recipes import DataFrame

from emir.dataproducts import MasterBias, MasterDark, MasterBadPixelMask
from emir.dataproducts import TelescopeOffset, MSMPositions
from emir.dataproducts import MasterIntensityFlat

from emir.dataproducts import DTU_XY_Calibration, DTU_Z_Calibration
from emir.dataproducts import CSU2DetectorCalibration, DTUFlexureCalibration
from emir.dataproducts import (PointingOriginCalibration, 
                             SpectroPhotometricCalibration)
from emir.dataproducts import PhotometricCalibration, WavelengthCalibration
__all__ = []

_logger = logging.getLogger('emir.recipes')

@provides(DTU_XY_Calibration)
class DTU_XY_CalibrationRecipe(RecipeBase):
    '''

    **Observing modes:**

        * DTU X_Y calibration
    '''

    __requires__ = [
        Parameter('slit_pattern', None, 'Slit pattern'),
        Parameter('dtu_range', None, 'DTU range: begin, end and step')
    ]

    def __init__(self):
        super(DTU_XY_CalibrationRecipe, self).__init__(
            author="Sergio Pascual <sergiopr@fis.ucm.es>",
            version="0.1.0"
        )

    def run(self, obresult):
        return {'products': [DTU_XY_Calibration()]}
    
@provides(DTU_Z_Calibration)
class DTU_Z_CalibrationRecipe(RecipeBase):
    '''

    **Observing modes:**

        * DTU Z calibration
    '''

    __requires__ = [
        Parameter('dtu_range', None, 'DTU range: begin, end and step')
    ]

    def __init__(self):
        super(DTU_Z_CalibrationRecipe, self).__init__(
            author="Sergio Pascual <sergiopr@fis.ucm.es>",
            version="0.1.0"
        )

    def run(self, obresult):
        return {'products': [DTU_Z_Calibration()]}

@requires()
@provides(DTUFlexureCalibration)
class DTUFlexureRecipe(RecipeBase):
    '''

    **Observing modes:**

        * DTU Flexure compensation
    '''

    # __requires__ = []

    def __init__(self):
        super(DTUFlexureRecipe, self).__init__(
            author="Sergio Pascual <sergiopr@fis.ucm.es>",
            version="0.1.0"
        )

    def run(self, obresult):
        return {'products': [DTUFlexureCalibration()]}

@provides(DTU_XY_Calibration)
class CSU2DetectorRecipe(RecipeBase):
    '''

    **Observing modes:**

        * CSU2Detector calibration
    '''

    __requires__ = [
        Parameter('dtu_range', None, 'DTU range: begin, end and step')
    ]

    def __init__(self):
        super(CSU2DetectorRecipe, self).__init__(
            author="Sergio Pascual <sergiopr@fis.ucm.es>",
            version="0.1.0"
        )

    def run(self, obresult):
        return {'products': [DTU_XY_Calibration()]}

@requires()
@provides(PointingOriginCalibration)
class FocalPlaneCalibrationRecipe(RecipeBase):
    '''

    **Observing modes:**

        * Lateral color
    '''

    # __requires__ = []

    def __init__(self):
        super(FocalPlaneCalibrationRecipe, self).__init__(
            author="Sergio Pascual <sergiopr@fis.ucm.es>",
            version="0.1.0"
        )

    def run(self, obresult):
        return {'products': [PointingOriginCalibration()]}

@requires()
@provides(WavelengthCalibration)
class SpectralCharacterizationRecipe(RecipeBase):
    '''

    **Observing modes:**

        * Spectral characterization
    '''

    def __init__(self):
        super(SpectralCharacterizationRecipe, self).__init__(
            author="Sergio Pascual <sergiopr@fis.ucm.es>",
            version="0.1.0"
        )

    def run(self, obresult):
        return {'products': [WavelengthCalibration()]}

@requires()
@provides(PointingOriginCalibration)
class RotationCenterRecipe(RecipeBase):
    '''

    **Observing modes:**

        * Centre of rotation
    '''

    # __requires__ = []

    def __init__(self):
        super(RotationCenterRecipe, self).__init__(
            author="Sergio Pascual <sergiopr@fis.ucm.es>",
            version="0.1.0"
        )

    def run(self, obresult):
        return {'products': [PointingOriginCalibration()]}

@requires()
@provides(DataFrame)
class AstrometricCalibrationRecipe(RecipeBase):
    '''

    **Observing modes:**

        * Astrometric calibration
    '''

    # __requires__ = []

    def __init__(self):
        super(AstrometricCalibrationRecipe, self).__init__(
            author="Sergio Pascual <sergiopr@fis.ucm.es>",
            version="0.1.0"
        )

    def run(self, obresult):
        return {'products': [DataFrame(None)]}

@provides(PhotometricCalibration)
class PhotometricCalibrationRecipe(RecipeBase):
    '''

    **Observing modes:**

        * Photometric calibration
    '''

    __requires__ = [
        Parameter('phot', None, 'Information about standard stars')
    ]

    def __init__(self):
        super(PhotometricCalibrationRecipe, self).__init__(
            author="Sergio Pascual <sergiopr@fis.ucm.es>",
            version="0.1.0"
        )

    def run(self, obresult):
        return {'products': [PhotometricCalibration()]}

@provides(SpectroPhotometricCalibration)
class SpectroPhotometricCalibrationRecipe(RecipeBase):
    '''

    **Observing modes:**

        * Spectrophotometric calibration
    '''

    __requires__ = [
        Parameter('sphot', None, 'Information about standard stars')
    ]

    def __init__(self):
        super(SpectroPhotometricCalibrationRecipe, self).__init__(
            author="Sergio Pascual <sergiopr@fis.ucm.es>",
            version="0.1.0"
        )

    def run(self, obresult):
        return {'products': [SpectroPhotometricCalibration()]}






#
# Copyright 2008-2012 Universidad Complutense de Madrid
# 
# This file is part of PyEmir
# 
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
# 

'''EMIR simulator.'''

from pkgutil import get_data
from StringIO import StringIO
import logging

import pyfits
import numpy
from numina.treedict import TreeDict
from numina.instrument import Shutter
from numina.instrument.template import interpolate

from emir.instrument.detector import EmirDetector, EmirDas

_logger = logging.getLogger('emir.simulator')

class BaseInstrument(object):
    def __init__(self, detector):

        self.shutter = Shutter()

        self.detector = detector
        self.das = EmirDas(detector)
        
        # Connect components
        self.detector.connect(self.shutter)
        
        self.detector.meta['gain'] = 2.8
        self.detector.meta['readnoise'] = 3.5

        self.meta = TreeDict()
        self.meta['name'] = 'EMIR'
        self.meta['focalstation'] = 'NASMYTH'
        self.meta['filter'] = 'J'
        self.meta['imagetype'] = ''
        self.meta['detector'] = self.detector.meta
        self.meta['das'] = self.das.meta
    
    def connect(self, source):
        self.shutter.connect(source)
 
    def filter(self, name):
        self.meta['filter'] = name

    def acquire(self, time):
        img = self.das.acquire(time)
        return self.meta, img

    def imagetype(self, name):
        self.meta['imagetype'] = name

class Instrument(BaseInstrument):
    def __init__(self):
        
        # FIXME: Testing flat
        xx, yy= numpy.mgrid[-1:1:2048j, -1:1:2048j]
        flat = 2 - 0.7 * (xx**2+yy**2)
        flat /= flat.mean()
        
        detector = EmirDetector(flat=flat)
        BaseInstrument.__init__(self, detector)

_result_types = ['image', 'spectrum']
_extensions = ['primary', 'variance', 'map', 'wcs']


_table = {('image','primary'): 'image_primary.txt',
          ('image', 'map'): 'image_map.txt',
          ('image', 'wcs'): 'image_wcs.txt',
          ('image', 'variance'): 'image_variance.txt',
          ('spectrum','primary'): 'spectrum_primary.txt',
          ('spectrum', 'map'): 'image_map.txt',
          ('spectrum', 'wcs'): 'image_wcs.txt',
          ('spectrum', 'variance'): 'image_variance.txt',
          }

def _load_header(res, ext):
    try:
        res = _table[(res, ext)]
    except KeyError:
        return pyfits.Header()    
    sfile = StringIO(get_data('emir.instrument', res))
    hh = pyfits.Header(txtfile=sfile)
    return hh

def _load_all_headers():
    result = {}
    for res in _result_types:
        result[res] = {}
        for ext in _extensions:
            result[res][ext] = _load_header(res, ext)
    
    return result
    
class EmirImageFactory(object):
    
    default = _load_all_headers() 

    # FIXME: workaround
    def __init__(self):
        sfile = StringIO(get_data('emir.instrument', 'image_primary.txt'))
        self.p_templ = pyfits.Header(txtfile=sfile)
        del sfile

    def create(self, meta, data):
        hh = self.p_templ.copy()
        
        for rr in hh.ascardlist():
            rr.value = interpolate(meta, rr.value)

        prim = pyfits.PrimaryHDU(data=data, header=hh)
        hl = [prim]

        hdulist = pyfits.HDUList(hl)
        return hdulist



#
# Copyright 2011-2012 Universidad Complutense de Madrid
# 
# This file is part of PyEmir
# 
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

'''
Image mode recipes of EMIR

'''

import logging

from numina.recipes import RecipeBase, Parameter, DataProductParameter
from numina.recipes import provides, DataFrame
from numina.recipes.requirements import Requirement

from emir.dataproducts import MasterBias, MasterDark, MasterBadPixelMask
from emir.dataproducts import MasterIntensityFlat
from emir.dataproducts import NonLinearityCalibration
from emir.dataproducts import SourcesCatalog

from .shared import DirectImageCommon

_logger = logging.getLogger('numina.recipes.emir')

@provides(DataFrame, SourcesCatalog)
class StareImageRecipe(RecipeBase, DirectImageCommon):
    '''
    The effect of recording images of the sky in a given pointing
    position of the TS


    **Observing modes:**

        * Stare image
    
    '''

    __requires__ = [
        DataProductParameter('master_bpm', MasterBadPixelMask, 
                  'Master bad pixel mask'),       
        DataProductParameter('master_bias', MasterBias, 'Master bias image'),
        DataProductParameter('master_dark', MasterDark, 'Master dark image'),
        DataProductParameter('nonlinearity', NonLinearityCalibration([1.0, 0.0]), 
                  'Polynomial for non-linearity correction'),
        DataProductParameter('master_intensity_ff', MasterIntensityFlat, 
                  'Master intensity flatfield'),
        Parameter('extinction', 0.0, 'Mean atmospheric extinction'),
        # FIXME: this parameter is optional 
        Parameter('sources', None, 
                  'List of x, y coordinates to measure FWHM',
                  optional=True),
        Parameter('offsets', None, 'List of pairs of offsets',
                  optional=True),
        Requirement('instrument.detector.channels', 'List of channels'),
        Requirement('instrument.detector.shape', 'Detector shape'),
    ]

    def __init__(self):
        super(StareImageRecipe, self).__init__(
            author="Sergio Pascual <sergiopr@fis.ucm.es>",
            version="0.1.0"
        )
        
    def run(self, obresult):
                
        baseshape = self.parameters['instrument.detector.shape']
        amplifiers = self.parameters['instrument.detector.channels']
        offsets = self.parameters['offsets']
        
        return self.process(obresult, baseshape, amplifiers, 
                            offsets=offsets, subpix=1,
                            stop_after=DirectImageCommon.PRERED)
        
        


#
# Copyright 2008-2012 Universidad Complutense de Madrid
# 
# This file is part of PyEmir
# 
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

'''Detector Nonlinearity Recipe.

Recipe to calibrate the detector nonlinearity.

**Observing modes:**

 * Not considered

**Inputs:**

 * A list of dome flat fields with the adequate exposure times.

**Outputs:**

 * The polynomial in a given format (TBD)

**Procedure:**

The median of the reduced flatfields is computed. For each exposure time, the
average and standard deviation of the ADUs is computed. A linear fit is
computed for the lowest ADU counts, with standard deviations as weights. A new
fit is performed, this time of ADU_linear/ADU_obs, versus ADU_obs. The
polynomial terms of the fit are computed.

'''




#
# Copyright 2011-2012 Universidad Complutense de Madrid
# 
# This file is part of PyEmir
# 
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

'''Auxiliary Recipes for EMIR'''

import logging

import numpy
import pyfits
from numina.recipes import RecipeBase, Parameter, provides, requires
from numina.recipes import DataProductParameter
from numina.logger import log_to_history
from numina.array.combine import median
from numina import __version__
from numina.flow import SerialFlow
from numina.flow.node import IdNode
from numina.flow.processing import BiasCorrector
from numina.flow.processing import DarkCorrector, NonLinearityCorrector, BadPixelCorrector

from ..dataproducts import MasterBias, MasterDark, MasterBadPixelMask
from ..dataproducts import NonLinearityCalibration, MasterIntensityFlat
from ..dataproducts import WavelengthCalibration, MasterSpectralFlat
from ..dataproducts import SlitTransmissionCalibration, ChannelLevelStatistics

__all__ = ['BiasRecipe', 'DarkRecipe', 'IntensityFlatRecipe',
           'SpectralFlatRecipe', 'SlitTransmissionRecipe',
           'WavelengthCalibrationRecipe']

_logger = logging.getLogger('emir.recipes')

_s_author = "Sergio Pascual <sergiopr@fis.ucm.es>"


@requires()
@provides(MasterBias, ChannelLevelStatistics)
class BiasRecipe(RecipeBase):
    '''
    Recipe to process data taken in Bias image Mode.

    Bias images only appear in Simple Readout mode.

    **Observing modes:**
    
     * Bias Image (3.1)   
    
    **Inputs:**
    
    **Outputs:**
    
     * A combined bias frame, with variance extension.
     * Statistics of the final image per channel (mean, median , variance) 
    
    **Procedure:**
    
    The list of images can be readly processed by combining them with a median algorithm.
    
    '''

    def __init__(self):
        super(BiasRecipe, self).__init__(author=_s_author,version="0.1.0")

    @log_to_history(_logger)
    def run(self, obresult):
        _logger.info('starting bias reduction')

        images = obresult.frames

        cdata = []

        try:
            for image in images:
                hdulist = pyfits.open(image, memmap=True, mode='readonly')
                cdata.append(hdulist)

            _logger.info('stacking %d images using median', len(cdata))
            
            data = median([d['primary'].data for d in cdata], dtype='float32')

            var2 = numpy.zeros_like(data[0])


            cls = ChannelLevelStatistics(exposure=0.0)

            for amp1, amp2 in self.instrument['amplifiers'][0]:
                
                region = (slice(*amp1), slice(*amp2))
                
                mean = numpy.mean(data[0][region])
                med = numpy.median(data[0][region])
                var = numpy.var(data[0][region])
                stts = mean, med, var
                sregion = amp1, amp2
                cls.statistics.append([sregion, stts])
                var2[region] = var

            hdu = pyfits.PrimaryHDU(data[0], header=cdata[0]['PRIMARY'].header)
    
            # update hdu header with
            # reduction keywords
            hdr = hdu.header
            hdr.update('FILENAME', 'master_bias-%(block_id)d.fits' % self.environ)
            hdr.update('IMGTYP', 'BIAS', 'Image type')
            hdr.update('NUMTYP', 'MASTER_BIAS', 'Data product type')
            hdr.update('NUMXVER', __version__, 'Numina package version')
            hdr.update('NUMRNAM', self.__class__.__name__, 'Numina recipe name')
            hdr.update('NUMRVER', self.__version__, 'Numina recipe version')

            exhdr = pyfits.Header()
            exhdr.update('extver', 1)
            varhdu = pyfits.ImageHDU(data[1], name='VARIANCE', header=exhdr)
            exhdr = pyfits.Header()
            exhdr.update('extver', 2)
            var2hdu = pyfits.ImageHDU(var2, name='VARIANCE', header=exhdr)
            num = pyfits.ImageHDU(data[2], name='MAP')

            hdulist = pyfits.HDUList([hdu, varhdu, var2hdu, num])

            _logger.info('bias reduction ended')

            return {'products': [MasterBias(hdulist), cls]}
        finally:
            for hdulist in cdata:
                hdulist.close()
            
@provides(MasterDark, ChannelLevelStatistics)   
class DarkRecipe(RecipeBase):
    '''Recipe to process data taken in Dark current image Mode.

    Recipe to process dark images. The dark images will be combined 
    using the median.
    They do have to be of the same exposure time t.
    
    **Observing mode:**
    
     * Dark current Image (3.2)
    
    **Inputs:**
    
    **Outputs:**
    
     * A combined dark frame, with variance extension.
    ''' 

    __requires__ = [DataProductParameter('master_bias', MasterBias, 
                              'Master bias calibration', optional=True)]

    def __init__(self):
        super(DarkRecipe, self).__init__(author=_s_author, version="0.1.0")
        
    @log_to_history(_logger)
    def run(self, obresult):
        _logger.info('starting dark reduction')

        cdata = []
        expdata = []

        try:
                        
            for name, exposure in obresult.images:
                hdulist = pyfits.open(name, memmap=True, mode='readonly')
                cdata.append(hdulist)
                expdata.append(exposure)

            if not all([exp == expdata[0] for exp in expdata]):
                _logger.error('image with wrong exposure time')
                raise RecipeError('image with wrong exposure time')

            _logger.info('stacking %d images using median', len(cdata))
            
            data = median([d['primary'].data for d in cdata], dtype='float32')
            hdu = pyfits.PrimaryHDU(data[0], header=cdata[0]['primary'].header)
            
        finally:
            for hdulist in cdata:
                hdulist.close()

        if self.parameters['master_bias'] is not None:
            # load bias
            
            master_bias = pyfits.open(self.parameters['master_bias'], mode='readonly')
            _logger.info('subtracting bias %s', str(self.parameters['master_bias']))
            # subtrac bias
            data[0] -= master_bias[0].data
            
            idx = master_bias.index_of(('variance', 1))
            data[1] += master_bias[idx].data
            

        var2 = numpy.zeros_like(data[0])

        cls = ChannelLevelStatistics(exposure=exposure)

        for amp1, amp2 in self.instrument['amplifiers'][0]:
            
            region = (slice(*amp1), slice(*amp2))
            
            mean = numpy.mean(data[0][region])
            med = numpy.median(data[0][region])
            var = numpy.var(data[0][region])
            stts = mean, med, var
            sregion = amp1, amp2
            cls.statistics.append([sregion, stts])
            var2[region] = var

        # update hdu header with
        # reduction keywords
        hdr = hdu.header
        hdr.update('NUMXVER', __version__, 'Numina package version')
        hdr.update('NUMRNAM', self.__class__.__name__, 'Numina recipe name')
        hdr.update('NUMRVER', self.__version__, 'Numina recipe version')
        
        hdr.update('FILENAME', 'master_dark-%(block_id)d.fits' % self.environ)
        hdr.update('IMGTYP', 'DARK', 'Image type')
        hdr.update('NUMTYP', 'MASTER_DARK', 'Data product type')
        
        exhdr = pyfits.Header()
        exhdr.update('extver', 1)
        varhdu = pyfits.ImageHDU(data[1], name='VARIANCE', header=exhdr)
        exhdr = pyfits.Header()
        exhdr.update('extver', 2)
        var2hdu = pyfits.ImageHDU(var2, name='VARIANCE', header=exhdr)
        num = pyfits.ImageHDU(data[2], name='MAP')

        hdulist = pyfits.HDUList([hdu, varhdu, var2hdu, num])

        md = MasterDark(hdulist)

        _logger.info('dark reduction ended')    

        return {'products': [md, cls]}


@provides(MasterIntensityFlat)
class IntensityFlatRecipe(RecipeBase):
    '''Recipe to process data taken in intensity flat-field mode.
        
    Recipe to process intensity flat-fields. The flat-on and flat-off images are
    combined (method?) separately and the subtracted to obtain a thermal subtracted
    flat-field.
    
    **Observing modes:**
    
     * Intensity Flat-Field
    
    **Inputs:**
    
      * A master dark frame
      * Non linearity
      * A model of the detector. 
    
    **Outputs:**
    
     * TBD
    
    **Procedure:**
    
     * A combined thermal subtracted flat field, normalized to median 1, 
       with with variance extension and quality flag. 
    
    '''
    __requires__ = [ 
        DataProductParameter('master_bias', MasterBias, 'Master bias image', optional=True),
        DataProductParameter('master_dark', MasterDark, 'Master dark image'),
        DataProductParameter('nonlinearity', NonLinearityCalibration([1.0, 0.0]), 
                  'Polynomial for non-linearity correction'),
    ]
    

    def __init__(self):
        super(IntensityFlatRecipe, self).__init__(
                        author=_s_author,
                        version="0.1.0"
                )
        
    @log_to_history(_logger)
    def run(self, obresult):
        _logger.info('starting flat reduction')
                
        # Basic processing
        if self.parameters['master_bias']:
            mbias = pyfits.getdata(self.parameters['master_bias'])
            bias_corrector = BiasCorrector(mbias)
        else:
            bias_corrector = IdNode()
            
        mdark = pyfits.getdata(self.parameters['master_dark'])
        dark_corrector = DarkCorrector(mdark)
        nl_corrector = NonLinearityCorrector(self.parameters['nonlinearity'])
        
        
        basicflow = SerialFlow([bias_corrector, dark_corrector, nl_corrector])

        for img in obresult.frames:

            with pyfits.open(img[0], mode='update') as hdulist:
                hdulist = basicflow(hdulist)

        # combine LAMP-ON
        
        lampon = [img[0] for img in obresult.frames if img[2] == 'LAMPON']
        
        dataon = self.stack_images(lampon)
                
        # combine LAMP-OFF
        lampoff = [img[0] for img in obresult.frames if img[2] == 'LAMPOFF']
        
        dataoff = self.stack_images(lampoff)
        
        # Subtract
        datafin = dataon[0] - dataoff[0]
        varfin = dataon[1] + dataoff[1]
        mapfin = dataon[2] + dataoff[2]

        meanval = datafin.mean()
        
        datafin /= meanval
        varfin /= (meanval * meanval)

        hdu = pyfits.PrimaryHDU(datafin)

        # update hdu header with
        # reduction keywords
        hdr = hdu.header
        hdr.update('NUMXVER', __version__, 'Numina package version')
        hdr.update('NUMRNAM', self.__class__.__name__, 'Numina recipe name')
        hdr.update('NUMRVER', self.__version__, 'Numina recipe version')
        
        hdr.update('FILENAME', 'master_intensity_flat-%(block_id)d.fits' % self.environ)
        hdr.update('IMGTYP', 'FLAT', 'Image type')
        hdr.update('NUMTYP', 'MASTER_FLAT', 'Data product type')
        
        varhdu = pyfits.ImageHDU(varfin, name='VARIANCE')
        num = pyfits.ImageHDU(mapfin, name='MAP')

        hdulist = pyfits.HDUList([hdu, varhdu, num])

        md = MasterIntensityFlat(hdulist)

        return {'products': [md]}
        
        
    def stack_images(self, images):
        
        cdata = []

        try:
            for name in images:
                hdulist = pyfits.open(name, memmap=True, mode='readonly')
                cdata.append(hdulist)

            _logger.info('stacking %d images using median', len(cdata))
            
            data = median([d['primary'].data for d in cdata], dtype='float32')
            return data
            
        finally:
            for hdulist in cdata:
                hdulist.close()
        
        
        
@provides(MasterSpectralFlat)
class SpectralFlatRecipe(RecipeBase):
    '''Spectral Flatfield Recipe.

    Recipe to process spectral flat-fields. The flat-on and flat-off images are
    combined (method?) separately and the subtracted to obtain a thermal subtracted
    flat-field.

    **Observing modes:**

        * Multislit mask Flat-Field
     
    **Inputs:**

     * A list of lamp-on flats
     * A model of the detector 

    **Outputs:**

     * A combined spectral flat field with with variance extension and quality flag.

    **Procedure:**

     * TBD

    '''


    __requires__ = [       
        DataProductParameter('master_bias', MasterBias, 'Master bias image'),
        DataProductParameter('master_dark', MasterDark, 'Master dark image'),
        DataProductParameter('master_bpm', MasterBadPixelMask, 'Master bad pixel mask'),
        DataProductParameter('nonlinearity', NonLinearityCalibration([1.0, 0.0]), 
                  'Polynomial for non-linearity correction'),
    ]

    def __init__(self):
        super(SpectralFlatRecipe, self).__init__(
            author="Sergio Pascual <sergiopr@fis.ucm.es>",
            version="0.1.0"
        )

    def run(self, obresult):
        return {'products': [MasterSpectralFlat(None)]}

@provides(SlitTransmissionCalibration)
class SlitTransmissionRecipe(RecipeBase):
    '''Recipe to calibrate the slit transmission.

    **Observing modes:**

        * Slit transmission calibration (4.4)
     
    **Inputs:**

        * A list of uniformly illuminated images of MSM

    **Outputs:**

     * A list of slit transmission functions 

    **Procedure:**

     * TBD

    '''
    

    __requires__ = [       
        DataProductParameter('master_bias', MasterBias, 'Master bias image'),
        DataProductParameter('master_dark', MasterDark, 'Master dark image'),
        DataProductParameter('master_bpm', MasterBadPixelMask, 'Master bad pixel mask'),
        DataProductParameter('nonlinearity', NonLinearityCalibration([1.0, 0.0]), 
                  'Polynomial for non-linearity correction'),
    ]

    def __init__(self):
        super(SlitTransmissionRecipe, self).__init__(
            author="Sergio Pascual <sergiopr@fis.ucm.es>",
            version="0.1.0"
        )

    @log_to_history(_logger)
    def run(self, obresult):
        return {'products': [SlitTransmissionCalibration()]}

        
        
@provides(WavelengthCalibration)
class WavelengthCalibrationRecipe(RecipeBase):
    '''Recipe to calibrate the spectral response.

    **Observing modes:**

        * Wavelength calibration (4.5)
     
    **Inputs:**

     * List of line positions
     * Calibrations upto spectral flatfielding

    **Outputs:**

     * Wavelength calibration structure 

    **Procedure:**

     * TBD
    '''

    __requires__ = [       
        DataProductParameter('master_bias', MasterBias, 'Master bias image'),
        DataProductParameter('master_dark', MasterDark, 'Master dark image'),
        DataProductParameter('master_bpm', MasterBadPixelMask, 'Master bad pixel mask'),
        DataProductParameter('nonlinearity', NonLinearityCalibration([1.0, 0.0]), 
                  'Polynomial for non-linearity correction'),
        DataProductParameter('master_intensity_ff', MasterIntensityFlat, 
                  'Master intensity flatfield'),
        DataProductParameter('master_spectral_ff', MasterSpectralFlat, 
                  'Master spectral flatfield'),
    ]

    def __init__(self):
        super(WavelengthCalibrationRecipe, self).__init__(
            author="Sergio Pascual <sergiopr@fis.ucm.es>",
            version="0.1.0"
        )

    @log_to_history(_logger)
    def run(self, obresult):
        return {'products': [WavelengthCalibration()]}




#
# Copyright 2012 Universidad Complutense de Madrid
# 
# This file is part of PyEmir
# 
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

'''
Image mode recipes of EMIR

'''

import logging

from numina.recipes import RecipeBase, Parameter, DataProductParameter
from numina.recipes import provides, DataFrame
from numina.recipes.requirements import Requirement

from emir.dataproducts import MasterBias, MasterDark, MasterBadPixelMask
from emir.dataproducts import MasterIntensityFlat
from emir.dataproducts import NonLinearityCalibration
from emir.dataproducts import SourcesCatalog

from .shared import DirectImageCommon

_logger = logging.getLogger('numina.recipes.emir')

@provides(DataFrame, SourcesCatalog)
class MicroditheredImageRecipe(RecipeBase, DirectImageCommon):
    '''
    Recipe for the reduction of microdithering imaging.
    
    The effect of recording a series of stare images, with the same
    acquisition parameters, and taken by pointing to a number of
    sky positions, with separations of the order of sub arcsecs,
    either by moving the either by nodding the TS, tilting the TS
    M2 or shifting the EMIR DTU, the latter being the most likely
    option. Displacements are of the order of fraction of pixels.
    Images share the large majority of the sky positions so they can
    be coadded. Used for improving the spatial resolution of the
    resulting images and not valid for sky or superflat images.
    

    **Observing modes:**

        * Micro-dithered images
    

    Recipe to reduce observations obtained in imaging mode with microdithering.
    A critical piece of information
    here is a table that clearly specifies which images can be labelled as
    *science*, and which ones as *sky*. Note that some images are used both as
    *science* and *sky* (when the size of the targets are small compared to the
    offsets).

    **Observing modes:**
     * Micro-dithered images 

    **Inputs:**


        * Offsets between them
        * Master Dark 
        * Bad pixel mask (BPM) 
        * Non-linearity correction polynomials 
        * Master flat (twilight/dome flats)
        * Master background (thermal background, only in K band)
        * Detector model (gain, RN, lecture mode)
        * Average extinction in the filter
        * Astrometric calibration (TBD)

    **Outputs:**

     * Image with three extensions: final image scaled to the individual exposure
       time, variance  and exposure time map OR number of images combined (TBD).

    **Procedure:**

    Images are regridded to a integer subdivision of the pixel and then they are
    corrected from dark, non-linearity and flat. It should be desirable that the
    microdithering follows a pattern that can be easily translated to a subdivision
    of the pixel size (by an integer *n* = 2, 3, 4,...) that does not requires a
    too high *n* value. An iterative process starts:

     * Sky is computed from each frame, using the list of sky images of each
       science frame. The objects are avoided using a mask (from the second
       iteration on).

     * The relatiev offsets are the nominal from the telescope. From the second
       iteration on, we refine them using bright objects.

     * We combine the sky-subtracted images, output is: a new image, a variance
       image and a exposure map/number of images used map.

     * An object mask is generated.

     * We recompute the sky map, using the object mask as an additional input. From
       here we iterate (typically 4 times).

     * Finally, the images are corrected from atmospheric extinction and flux
       calibrated.

     * A preliminary astrometric calibration can always be used (using the central
       coordinates of the pointing and the plate scale in the detector). A better
       calibration might be computed using available stars (TBD).

    '''


    logger = _logger

    __requires__ = [
        DataProductParameter('master_bpm', MasterBadPixelMask, 
                  'Master bad pixel mask'),       
        DataProductParameter('master_bias', MasterBias, 'Master bias image', 
					optional=True),
        DataProductParameter('master_dark', MasterDark, 'Master dark image'),
        DataProductParameter('nonlinearity', NonLinearityCalibration([1.0, 0.0]), 
                  'Polynomial for non-linearity correction'),
        DataProductParameter('master_intensity_ff', MasterIntensityFlat, 
                  'Master intensity flatfield'),
        Parameter('extinction', 0.0, 'Mean atmospheric extinction'), 
        Parameter('sources', None, 
                  'List of x, y coordinates to measure FWHM',
                  optional=True),
        Parameter('subpixelization', 4, 
                  'Number of subdivisions in each pixel side'),
        Parameter('window', None, 'Region of interesting data', optional=True),
        Parameter('offsets', None, 'List of pairs of offsets',
                  optional=True),
        Requirement('instrument.detector.channels', 'List of channels'),
        Requirement('instrument.detector.shape', 'Detector shape'),
        Parameter('check_photometry_levels', [0.5, 0.8], 'Levels to check the flux of the objects'),
        Parameter('check_photometry_actions', ['warn', 'warn', 'default'], 'Actions to take on images'),
    ]

    def __init__(self):
        super(MicroditheredImageRecipe, self).__init__(
            author="Sergio Pascual <sergiopr@fis.ucm.es>",
            version="0.1.0"
        )
        
    def run(self, obresult):
        
        subpix = self.parameters['subpixelization']
        baseshape = self.parameters['instrument.detector.shape']
        amplifiers = self.parameters['instrument.detector.channels']
        offsets = self.parameters['offsets']
        window = self.parameters['window']
        offsets = self.parameters['offsets']
        
        return self.process(obresult, baseshape, amplifiers, window=window, 
                            offsets=offsets, subpix=subpix,
                            stop_after=DirectImageCommon.FULLRED)
        

    



#
# Copyright 2008-2012 Universidad Complutense de Madrid
# 
# This file is part of PyEmir
# 
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
# 

'''Recipe for image simulation.'''


import logging

import numpy

from numina.recipes import RecipeBase
import numina.qa as qa
from numina.simulation import run_counter

from emir.instrument.detector import Hawaii2Detector
from emir.dataproducts import create_raw

_logger = logging.getLogger("emir.recipes")

class Recipe(RecipeBase):
    '''Recipe to simulate EMIR images.'''
    
    __requires__ = [
#        Schema('detector', ProxyQuery(dummy={}), 'Detector parameters'),
#        Schema('readout', ProxyQuery(dummy={}), 'Readout mode'),
#        Schema('name_format', 'r%05d', 'Filename format'),
    ]

    capabilities = ['simulate_image']
    
    def __init__(self, parameters, runinfo):
        super(Recipe, self).__init__(parameters, runinfo)
        #
        _logger.info('Run counter created')
        self.runcounter = run_counter("r%05d", suffix='')
        
        _logger.info('Creating detector')
        
        self.detector = Hawaii2Detector(**self.parameters['detector'])
        _logger.info('Configuring detector')
        self.detector.configure(self.parameters['readout'])
        
        self.input = numpy.zeros(self.parameters['detector']['shape'])
        self.detector.exposure(self.parameters['readout']['exposure'])
        self.repeat = self.parameters['readout']['repeat']
     
    def run(self):
        _logger.info('Creating simulated array')    
        output = self.detector.lpath(self.input)
        run = self.runcounter.next()
        headers = {'RUN': run, 'FILENAME': '%s.fits' % run}
        
        _logger.info('Collecting detector metadata')
        headers.update(self.detector.metadata())
        
        _logger.info('Building FITS structure')
        hdulist = create_raw(output, headers)
        return {'qa': qa.UNKNOWN, 'simulated_image': hdulist}



#
# Copyright 2010-2012 Universidad Complutense de Madrid
# 
# This file is part of PyEmir
# 
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
# 

'''Recipe for the reordering of frames.'''

import logging
import os.path
import itertools as ito

import numpy # pylint: disable-msgs=E1101


import numina.qa
from numina.recipes import RecipeBase

from emir.instrument.detector import CHANNELS_READOUT
from emir.dataproducts import create_result

_logger = logging.getLogger("emir.recipes")

class Recipe(RecipeBase):
    '''Reordering Recipe.
    
    Recipe to reorder images created by the detector.
    '''
    required_parameters = [
    ]
    
    
    def __init__(self, param, runinfo):
        super(Recipe, self).__init__(param, runinfo)
        
    def setup(self):
        self.parameters['images'] = [DiskImage(os.path.abspath(path)) 
                  for path in self.parameters['images']]

    def run(self):
        
        images = self.parameters['images']
        
        results = []

        for img in images:
            # Using numpy memmap instead of pyfits
            # Images are a in a format unrecognized by pyfits
            _logger.debug('processing %s', img.filename)
            f = numpy.memmap(img.filename, 
                                  dtype='uint16', mode='r', offset=36 * 80)            
            try:                                
                f.shape = (1024, 4096)
                rr = numpy.zeros((2048, 2048), dtype='int16')
                for idx, channel, conv in ito.izip(ito.count(0), 
                                                   CHANNELS_READOUT,
                                                   ito.cycle([lambda x:x, lambda x:x.T])
                                                   ):
                    rr[channel] = conv(f[:,slice(128 * idx, 128 * (idx + 1))])

                basename = os.path.basename(img.filename)
                primary_headers = {'FILENAME': basename}
                result = create_result(rr, 
                                       headers=primary_headers)
                result.writeto(basename)
                
                result.close()                
                results.append(DiskImage(os.path.abspath(basename)))
            finally:
                del f
                img.close()
        
        return {'qa': numina.qa.UNKNOWN, 'images': results}
                                                  
    
if __name__ == '__main__':
    import uuid
    import glob
    
    from numina.compatibility import json
    from numina.user import main
    from numina.jsonserializer import to_json

    logging.basicConfig(level=logging.DEBUG)
    _logger.setLevel(logging.DEBUG)
        
    pv = {'recipes': {'default': {'parameters': {
                                    'region': 'channel'
                                },
                     'run': {'repeat': 1,
                             },   
                        },                 
                      },                      
          'observing_block': {'instrument': 'emir',
                       'mode': 'detector_reorder',
                       'id': 1,
                       'result': {
                                  'images': [], 

                            },
                       },                     
    }
    os.chdir('/home/spr/Datos/emir/test6/data')
    
    pv['observing_block']['result']['images'] = glob.glob('*serie*.fits')
    #pv['observing_block']['result']['images'] = ['pozo_BG33_VR05_OFF11_serie.0054.fits']
    os.chdir('/home/spr/Datos/emir/test6')
    
    # Creating base directory for storing results
    uuidstr = str(uuid.uuid1()) 
    basedir = os.path.join(os.getcwd(), uuidstr)
    os.mkdir(basedir)
    
    ff = open('config.txt', 'w+')
    try:
        json.dump(pv, ff, default=to_json, encoding='utf-8', indent=2)
    finally:
        ff.close()

    main(['-d', '--basedir', basedir, '--datadir', 'data',
          '--run', 'config.txt'])



#
# Copyright 2008-2012 Universidad Complutense de Madrid
# 
# This file is part of PyEmir
# 
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

'''
Recipe for the processing of target acquisition images.

**Observing modes:**

    * Target acquisition
'''

import logging

from numina.recipes import RecipeBase, Parameter, DataProductParameter
from numina.recipes import provides

from ..dataproducts import MasterBias, MasterDark, MasterBadPixelMask
from ..dataproducts import TelescopeOffset, MSMPositions
from ..dataproducts import MasterIntensityFlat
from ..dataproducts import NonLinearityCalibration

__all__ = ['TargetAcquisitionRecipe', 'MaskImagingRecipe', 'MaskCheckRecipe']

_logger = logging.getLogger('emir.recipes')

@provides(TelescopeOffset)
class TargetAcquisitionRecipe(RecipeBase):
    '''
    Acquire a target.
    
    Recipe for the processing of target acquisition images.

    **Observing modes:**

        * Target acquisition
    
    
    '''

    __requires__ = [       
        DataProductParameter('master_bias', MasterBias, 'Master bias image'),
        DataProductParameter('master_dark', MasterDark, 'Master dark image'),
        DataProductParameter('master_bpm', MasterBadPixelMask, 'Master bad pixel mask'),
        DataProductParameter('nonlinearity', NonLinearityCalibration([1.0, 0.0]), 
                  'Polynomial for non-linearity correction'),
        DataProductParameter('master_intensity_ff', MasterIntensityFlat, 
                  'Master intensity flatfield'),
    ]

    def __init__(self):
        super(TargetAcquisitionRecipe, self).__init__(
            author="Sergio Pascual <sergiopr@fis.ucm.es>",
            version="0.1.0"
        )

    def run(self, obresult):
        return {'products': [TelescopeOffset()]}
    
@provides(MSMPositions)
class MaskImagingRecipe(RecipeBase):
    '''Acquire a target.
    
    Mask image Recipe.

    Recipe to process mask images.

    **Observing modes:**

      *  Mask imaging
    '''

    __requires__ = [       
        DataProductParameter('master_bias', MasterBias, 'Master bias image'),
        DataProductParameter('master_dark', MasterDark, 'Master dark image'),
        DataProductParameter('master_bpm', MasterBadPixelMask, 'Master bad pixel mask'),
        DataProductParameter('nonlinearity', NonLinearityCalibration([1.0, 0.0]), 
                  'Polynomial for non-linearity correction'),
        DataProductParameter('master_intensity_ff', MasterIntensityFlat, 
                  'Master intensity flatfield'),
    ]

    def __init__(self):
        super(MaskImagingRecipe, self).__init__(
            author="Sergio Pascual <sergiopr@fis.ucm.es>",
            version="0.1.0"
        )

    def run(self, obresult):
        return {'products': [MSMPositions()]}
    
@provides(TelescopeOffset, MSMPositions)
class MaskCheckRecipe(RecipeBase):
    '''
    Acquire a target.
    
    Recipe for the processing of multi-slit/long-slit check images.

    **Observing modes:**

        * MSM and LSM check 
            
    '''

    __requires__ = [       
        DataProductParameter('master_bias', MasterBias, 'Master bias image'),
        DataProductParameter('master_dark', MasterDark, 'Master dark image'),
        DataProductParameter('master_bpm', MasterBadPixelMask, 'Master bad pixel mask'),
        DataProductParameter('nonlinearity', NonLinearityCalibration([1.0, 0.0]), 
                  'Polynomial for non-linearity correction'),
        DataProductParameter('master_intensity_ff', MasterIntensityFlat, 
                  'Master intensity flatfield'),
    ]

    def __init__(self):
        super(MaskCheckRecipe, self).__init__(
            author="Sergio Pascual <sergiopr@fis.ucm.es>",
            version="0.1.0"
        )

    def run(self, obresult):
        return {'products': [TelescopeOffset(), MSMPositions()]}


#
# Copyright 2011-2012 Universidad Complutense de Madrid
# 
# This file is part of PyEmir
# 
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

'''

Routines shared by image mode recipes

'''

import logging
import operator

import numpy
import pyfits
from numina.util.sextractor import SExtractor

from .naming import name_skysub_proc

_logger = logging.getLogger('numina.recipes.emir')

# Actions to carry over images when checking the flux
# of the objects in different images
def warn_action(img):
    _logger.warn('Image %s has low flux in objects', img.baselabel)
    img.valid_science = True

def reject_action(img):
    img.valid_science = False
    _logger.info('Image %s rejected, has low flux in objects', img.baselabel)            
    pass

def default_action(img):
    _logger.info('Image %s accepted, has correct flux in objects', img.baselabel)      
    img.valid_science = True

# Actions
_dactions = {'warn': warn_action, 'reject': reject_action, 'default': default_action}

def check_photometry(frames, sf_data, seeing_fwhm, step=0, border=300, extinction=0.0,
                     check_photometry_levels=[0.5, 0.8], check_photometry_actions=['warn', 'warn', 'default'], figure=None):
    # Check photometry of few objects
    weigthmap = 'weights4rms.fits'
    
    wmap = numpy.zeros_like(sf_data[0])
    
    # Center of the image
    wmap[border:-border, border:-border] = 1                    
    pyfits.writeto(weigthmap, wmap, clobber=True)
    
    basename = 'result_i%0d.fits' % (step)
    sex = SExtractor()
    sex.config['VERBOSE_TYPE'] = 'QUIET'
    sex.config['PIXEL_SCALE'] = 1
    sex.config['BACK_TYPE'] = 'AUTO' 
    if seeing_fwhm is not None:
        sex.config['SEEING_FWHM'] = seeing_fwhm * sex.config['PIXEL_SCALE']
    sex.config['WEIGHT_TYPE'] = 'MAP_WEIGHT'
    sex.config['WEIGHT_IMAGE'] = weigthmap
    
    sex.config['PARAMETERS_LIST'].append('FLUX_BEST')
    sex.config['PARAMETERS_LIST'].append('FLUXERR_BEST')
    sex.config['PARAMETERS_LIST'].append('FWHM_IMAGE')
    sex.config['PARAMETERS_LIST'].append('CLASS_STAR')
    
    sex.config['CATALOG_NAME'] = 'master-catalogue-i%01d.cat' % step
    _logger.info('Runing sextractor in %s', basename)
    sex.run('%s,%s' % (basename, basename))
    
    # Sort catalog by flux
    catalog = sex.catalog()
    catalog = sorted(catalog, key=operator.itemgetter('FLUX_BEST'), reverse=True)
    
    # set of indices of the N first objects
    OBJS_I_KEEP = 3
    indices = set(obj['NUMBER'] for obj in catalog[:OBJS_I_KEEP])
    
    base = numpy.empty((len(frames), OBJS_I_KEEP))
    error = numpy.empty((len(frames), OBJS_I_KEEP))
    
    for idx, frame in enumerate(frames):
        imagename = name_skysub_proc(frame.baselabel, step)

        sex.config['CATALOG_NAME'] = 'catalogue-%s-i%01d.cat' % (frame.baselabel, step)

        # Lauch SExtractor on a FITS file
        # om double image mode
        _logger.info('Runing sextractor in %s', imagename)
        sex.run('%s,%s' % (basename, imagename))
        catalog = sex.catalog()
        
        # Extinction correction
        excor = pow(10, -0.4 * frame.airmass * extinction)
        base[idx] = [obj['FLUX_BEST'] / excor
                                 for obj in catalog if obj['NUMBER'] in indices]
        error[idx] = [obj['FLUXERR_BEST'] / excor
                                 for obj in catalog if obj['NUMBER'] in indices]
    
    data = base / base[0]
    err = error / base[0] # sigma
    w = 1 / err / err
    # weighted mean of the flux values
    wdata = numpy.average(data, axis=1, weights=w)
    wsigma = 1 / numpy.sqrt(w.sum(axis=1))
    
    levels = check_photometry_levels
    actions = check_photometry_actions
    
    x = range(len(frames))
    vals, (_, sigma) = check_photometry_categorize(x, wdata, levels, tags=actions)
    # n sigma level to plt
    n = 3
    
    if figure is not None:
        check_photometry_plot(figure, vals, wsigma, levels, n * sigma)
    
    for x, _, t in vals:
        try:
            action = _dactions[t]
        except KeyError:
            _logger.warning('Action named %s not recognized, ignoring', t)
            action = default_action
        for p in x:
            action(frames[p])
            
def check_photometry_categorize(x, y, levels, tags=None):
    '''Put every point in its category.

    levels must be sorted.'''   
    x = numpy.asarray(x)
    y = numpy.asarray(y)
    ys = y.copy()
    ys.sort()
    # Mean of the upper half
    m = ys[len(ys) / 2:].mean()
    y /= m
    m = 1.0
    s = ys[len(ys) / 2:].std()
    result = []

    if tags is None:
        tags = range(len(levels) + 1)

    for l, t in zip(levels, tags):
        indc = y < l
        if indc.any():
            x1 = x[indc]
            y1 = y[indc]
            result.append((x1, y1, t))

            x = x[indc == False]
            y = y[indc == False]
    else:
        result.append((x, y, tags[-1]))

    return result, (m, s)
           
def check_photometry_plot(figure, vals, errors, levels, nsigma, step=0):
    x = range(len(errors))
    figure.clf()
    ax = figure.add_subplot(111)
    ax.set_title('Relative flux of brightest object')
    for v,c in zip(vals, ['b', 'r', 'g', 'y']):
        ax.scatter(v[0], v[1], c=c)
        w = errors[v[0]]
        ax.errorbar(v[0], v[1], yerr=w, fmt=None, c=c)
        

    ax.plot([x[0], x[-1]], [1, 1], 'r--')
    ax.plot([x[0], x[-1]], [1 - nsigma, 1 - nsigma], 'b--')
    for f in levels:
        ax.plot([x[0], x[-1]], [f, f], 'g--')
        
    figure.canvas.draw()
    figure.savefig('figure-relative-flux_i%01d.png' % step)

def check_position(images_info, sf_data, seeing_fwhm, step=0):
        # FIXME: this method has to be updated

        _logger.info('Checking positions')
        # Check position of bright objects
        weigthmap = 'weights4rms.fits'
        
        wmap = numpy.zeros_like(sf_data[0])
        
        # Center of the image
        border = 300
        wmap[border:-border, border:-border] = 1                    
        pyfits.writeto(weigthmap, wmap, clobber=True)
        
        basename = 'result_i%0d.fits' % (step)
        sex = SExtractor()
        sex.config['VERBOSE_TYPE'] = 'QUIET'
        sex.config['PIXEL_SCALE'] = 1
        sex.config['BACK_TYPE'] = 'AUTO'
        if  seeing_fwhm is not None and seeing_fwhm > 0:
            sex.config['SEEING_FWHM'] = seeing_fwhm * sex.config['PIXEL_SCALE']
        sex.config['WEIGHT_TYPE'] = 'MAP_WEIGHT'
        sex.config['WEIGHT_IMAGE'] = weigthmap
        
        sex.config['PARAMETERS_LIST'].append('FLUX_BEST')
        sex.config['PARAMETERS_LIST'].append('FLUXERR_BEST')
        sex.config['PARAMETERS_LIST'].append('FWHM_IMAGE')
        sex.config['PARAMETERS_LIST'].append('CLASS_STAR')
        
        sex.config['CATALOG_NAME'] = 'master-catalogue-i%01d.cat' % step
        
        _logger.info('Runing sextractor in %s', basename)
        sex.run('%s,%s' % (basename, basename))
        
        # Sort catalog by flux
        catalog = sex.catalog()
        catalog = sorted(catalog, key=operator.itemgetter('FLUX_BEST'), reverse=True)
        
        # set of indices of the N first objects
        OBJS_I_KEEP = 10
        
        # master = [(obj['X_IMAGE'], obj['Y_IMAGE']) for obj in catalog[:OBJS_I_KEEP]]
        
        for image in images_info:
            imagename = name_skysub_proc(image.baselabel, step)

            sex.config['CATALOG_NAME'] = 'catalogue-self-%s-i%01d.cat' % (image.baselabel, step)

            # Lauch SExtractor on a FITS file
            # on double image mode
            _logger.info('Runing sextractor in %s', imagename)
            sex.run(imagename)
            catalog = sex.catalog()
            
            
            # data = [(obj['X_IMAGE'], obj['Y_IMAGE']) for obj in catalog]
            
            #tree = KDTree(data)
            
            # Search 2 neighbors
            #dists, _ids = tree.query(master, 2, distance_upper_bound=5)
            
            #for i in dists[:,0]:
            #    print i
            
            
            #_logger.info('Mean offset correction for image %s is %f', imagename, dists[:,0].mean())
            #raw_input('press any key')



#
# Copyright 2008-2012 Universidad Complutense de Madrid
# 
# This file is part of PyEmir
# 
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
# 

'''The EMIR Data Reduction Pipeline'''

import logging

from numina.pipeline import register_pipeline

from emir.recipes import EmirPipeline
from emir import __version__

register_pipeline(EmirPipeline(version=__version__))



#
# Copyright 2011-2012 Universidad Complutense de Madrid
# 
# This file is part of PyEmir
# 
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

'''

Routines shared by image mode recipes

'''
import os
import logging
import shutil
import math

import numpy
import pyfits
import pywcs
from scipy.spatial import KDTree as KDTree
import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.patches import Ellipse
from matplotlib.collections import PatchCollection
from numina import __version__
from numina.recipes import DataFrame
from numina.flow import SerialFlow 
from numina.flow.node import IdNode
from numina.flow.processing import BiasCorrector, FlatFieldCorrector
from numina.flow.processing import DarkCorrector, NonLinearityCorrector, BadPixelCorrector
from numina.array import fixpix2
from numina.frame import resize_fits, custom_region_to_str
from numina.array import combine_shape, correct_flatfield
from numina.array import subarray_match
from numina.array.combine import flatcombine, median, quantileclip

from numina.util.sextractor import SExtractor
from numina.util.sextractor import open as sopen
import numina.util.sexcatalog as sexcatalog

from emir.dataproducts import SourcesCatalog

from .checks import check_photometry
from .naming import name_redimensioned_frames, name_object_mask, name_skybackground
from .naming import name_skybackgroundmask, name_skysub_proc, name_skyflat
from .naming import name_skyflat_proc, name_segmask


_logger = logging.getLogger('numina.recipes.emir')

def offsets_from_wcs(frames, pixref):
    '''Compute offsets between frames using WCS information.
    
    :parameter frames: sequence of FITS filenames or file descriptors
    :parameter pixref: numpy array used as reference pixel
    
    The sky world coordinates are computed on *pixref* using
    the WCS of the first frame in the sequence. Then, the
    pixel coordinates of the reference sky world-coordinates 
    are computed for the rest of the frames.
    
    The results is a numpy array with the difference between the
    computed pixel value and the reference pixel. The first line
    of the array is [0, 0], being the offset from the first image
    to itself. 
    
    '''
    
    result = numpy.zeros((len(frames), pixref.shape[1]))

    with pyfits.open(frames[0]) as hdulist:
        wcs = pywcs.WCS(hdulist[0].header)
        skyref = wcs.wcs_pix2sky(pixref, 1)

    for idx, frame in enumerate(frames[1:]):
        with pyfits.open(frame) as hdulist:
            wcs = pywcs.WCS(hdulist[0].header)
            pixval = wcs.wcs_sky2pix(skyref, 1)
            result[idx + 1] = pixval[0] - pixref[0]

    return result

def intersection(a, b, scale=1):
    '''Intersection between two segments.'''
    a1, a2 = a
    try:
        b1, b2 = b
    except TypeError:
        b1 = b.start
        b2 = b.stop
        

    if a2 <= b1:
        return None
    if a1 >= b2:
        return None

    # a2 > b1 and a1 < b2

    if a2 <= b2:
        if a1 <= b1:
            return slice(b1 * scale, a2 * scale)
        else:
            return slice(a1 * scale, a2 * scale)
    else:
        if a1 <= b1:
            return slice(b1 * scale, b2 * scale)
        else:
            return slice(a1 * scale, b2 * scale)

def clip_slices(r, region, scale=1):
    '''Intersect slices with a region.''' 
    t = []
    for ch in r:
        a1 = intersection(ch[0], region[0], scale=scale)
        if a1 is None:
            continue
        a2 = intersection(ch[1], region[1], scale=scale)
        if a2 is None:
            continue

        t.append((a1, a2))

    return t

class DirectImageCommon(object):
        
    logger = _logger
    BASIC, PRERED, CHECKRED, FULLRED, COMPLETE = range(5)
    
    def __init__(self, *args, **kwds):
        super(DirectImageCommon, self).__init__()
        
        self._figure = plt.figure(facecolor='white')
        self._figure.canvas.set_window_title('Recipe Plots')
        self._figure.canvas.draw()
    
    def process(self, obresult, baseshape, channels, 
                offsets=None, window=None, 
                subpix=1, store_intermediate=True,
                target_is_sky=True, stop_after=PRERED):
        
        numpy.seterr(divide='raise')
        
        # metadata = self.instrument['metadata']
        # FIXME: hardcoded
        keywords = {
         "juliandate": "MJD-OBS", 
         "airmass": "AIRMASS", 
         "imagetype": "IMGTYP", 
         "exposure": "EXPTIME"
        }     
        recipe_result = {'products' : []}

        if window is None:
            window = tuple((0, siz) for siz in baseshape)

        if store_intermediate:
            recipe_result['intermediate'] = []
        
        
        # States        
        state = self.BASIC
        step = 0
        
        try:
            niteration = self.parameters['iterations']
        except KeyError:
            niteration = 1
        
        
        while True:
            if state == self.BASIC:    
                _logger.info('Basic processing')

                # Basic processing
                
                # FIXME: add this
                bpm = pyfits.getdata(self.parameters['master_bpm'])
                bpm_corrector = BadPixelCorrector(bpm)
                
                if self.parameters['master_bias']:
                    mbias = pyfits.getdata(self.parameters['master_bias'])
                    bias_corrector = BiasCorrector(mbias)
                else:
                    bias_corrector = IdNode()
            
                mdark = pyfits.getdata(self.parameters['master_dark'])
                dark_corrector = DarkCorrector(mdark)
                nl_corrector = NonLinearityCorrector(self.parameters['nonlinearity'])

                mflat = pyfits.getdata(self.parameters['master_intensity_ff'])
                ff_corrector = FlatFieldCorrector(mflat)  
                  
                basicflow = SerialFlow([bpm_corrector,
                                        bias_corrector, 
                                        dark_corrector, 
                                        nl_corrector,
                                        ff_corrector
                                        ])

                for frame in obresult.frames:
                    with pyfits.open(frame.label, mode='update') as hdulist:
                            hdulist = basicflow(hdulist)
                  
                if stop_after == state:
                    break
                else:
                    state = self.PRERED
            elif state == self.PRERED:                
                # Shape of the window
                windowshape = tuple((i[1] - i[0]) for i in window)
                _logger.debug('Shape of window is %s', windowshape)
                # Shape of the scaled window
                subpixshape = tuple((side * subpix) for side in windowshape)
                    
                # Scaled window region
                scalewindow = tuple(slice(*(subpix * i for i in p)) for p in window)
                # Window region
                window = tuple(slice(*p) for p in window)
                
                scaled_chan = clip_slices(channels, window, scale=subpix)

                # Reference pixel in the center of the frame
                refpix = numpy.divide(numpy.array([baseshape], dtype='int'), 2).astype('float')
        
                # lists of targets and sky frames
                targetframes = []
                skyframes = []
                
                for frame in obresult.frames:
                    
                    # Getting some metadata from FITS header
                    hdr = pyfits.getheader(frame.label)
                    try:
                        frame.exposure = hdr[str(keywords['exposure'])]
                        #frame.baseshape = get_image_shape(hdr)
                        frame.airmass = hdr[str(keywords['airmass'])]
                        frame.mjd = hdr[str(keywords['juliandate'])]
                    except KeyError as e:
                        raise KeyError("%s in frame %s" % (str(e), frame.label))
                    
                    
                    frame.baselabel = os.path.splitext(frame.label)[0]
                    frame.mask = self.parameters['master_bpm']
                    # Insert pixel offsets between frames    
                    frame.objmask_data = None
                    frame.valid_target = False
                    frame.valid_sky = False
                    frame.valid_region = scalewindow
                    if frame.itype == 'TARGET':
                        frame.valid_target = True
                        targetframes.append(frame)
                        if target_is_sky:
                            frame.valid_sky = True
                            skyframes.append(frame)
                    if frame.itype == 'SKY':
                        frame.valid_sky = True
                        skyframes.append(frame)
        
                labels = [frame.label for frame in targetframes]
        
                if offsets is None:
                    _logger.info('Computing offsets from WCS information')
                    
                    list_of_offsets = offsets_from_wcs(labels, refpix)
                else:
                    _logger.info('Using offsets from parameters')
                    list_of_offsets = numpy.asarray(offsets)

                # Insert pixel offsets between frames
                for frame, off in zip(targetframes, list_of_offsets):
                    
                    # Insert pixel offsets between frames
                    frame.pix_offset = off
                    frame.scaled_pix_offset = subpix * off
 
                    _logger.debug('Frame %s, offset=%s, scaled=%s', frame.label, off, subpix * off)

                _logger.info('Computing relative offsets')
                offsets = [(frame.scaled_pix_offset) 
                           for frame in targetframes]
                offsets = numpy.round(offsets).astype('int')        
                finalshape, offsetsp = combine_shape(subpixshape, offsets)
                _logger.info('Shape of resized array is %s', finalshape)
                
                # Resizing target frames        
                self.resize(targetframes, subpixshape, offsetsp, finalshape, 
                            window=window, scale=subpix)
                
                if not target_is_sky:
                    for frame in skyframes:
                        frame.resized_base = frame.label
                        frame.resized_mask = frame.mask    
                
                # superflat
                _logger.info('Step %d, superflat correction (SF)', step)
                # Compute scale factors (median)           
                self.update_scale_factors(obresult.frames)

                # Create superflat                
                superflat = self.compute_superflat(skyframes, channels=scaled_chan,
                                                   target_is_sky=target_is_sky, step=step)
            
                # Apply superflat
                self.figure_init(subpixshape)
                self.apply_superflat(obresult.frames, superflat)

                _logger.info('Simple sky correction')
                if target_is_sky:
                    # Each frame is the closest sky frame available
                    
                    for frame in obresult.frames:            
                        self.compute_simple_sky_for_frame(frame, frame)
                else:
                    self.compute_simple_sky(targetframes, skyframes)
                
                # Combining the frames
                _logger.info("Step %d, Combining target frames", step)
                
                sf_data = self.combine_frames(targetframes)
                    
                self.figures_after_combine(sf_data)
                      
                _logger.info('Step %d, finished', step)

                if stop_after == state:
                    break
                else:
                    state = self.CHECKRED
            elif state == self.CHECKRED:
                
                seeing_fwhm = None

                #self.check_position(images_info, sf_data, seeing_fwhm)
                recompute = False
                if recompute:
                    _logger.info('Recentering is needed')
                    state = self.PRERED
                else:
                    _logger.info('Recentering is not needed')
                    _logger.info('Checking photometry')
                    check_photometry(targetframes, sf_data, seeing_fwhm, figure=self._figure)
                    
                    if stop_after == state:
                        break
                    else:
                        state = self.FULLRED
            elif state == self.FULLRED:

                # Generating segmentation image
                _logger.info('Step %d, generating segmentation image', step)
                objmask, seeing_fwhm = self.create_mask(sf_data, seeing_fwhm, step=step)
                step +=1
                # Update objects mask
                # For all images    
                # FIXME:
                for frame in targetframes:
                    frame.objmask = name_object_mask(frame.baselabel, step)
                    _logger.info('Step %d, create object mask %s', step,  frame.objmask)                 
                    frame.objmask_data = objmask[frame.valid_region]
                    pyfits.writeto(frame.objmask, frame.objmask_data, clobber=True)
                    
                if not target_is_sky:
                    
                    for frame in skyframes:
                        pass

                _logger.info('Step %d, superflat correction (SF)', step)
                
                # Compute scale factors (median)           
                self.update_scale_factors(obresult.frames, step)

                # Create superflat
                superflat = self.compute_superflat(obresult.frames, scaled_chan, target_is_sky,
                                                   segmask=objmask, step=step)
                
                # Apply superflat
                self.figure_init(subpixshape)
                
                self.apply_superflat(obresult.frames, superflat, step=step, save=True)

                _logger.info('Step %d, advanced sky correction (SC)', step)                
                self.compute_advanced_sky(targetframes, objmask, 
                                          skyframes=skyframes,
                                          target_is_sky=target_is_sky,
                                          step=step)
            
                # Combining the images
                _logger.info("Step %d, Combining the images", step)
                # FIXME: only for science
                sf_data = self.combine_frames(targetframes, step=step)
                self.figures_after_combine(sf_data)

                if step >= niteration:
                    state = self.COMPLETE
            else:
                break

        hdu = pyfits.PrimaryHDU(sf_data[0])                
        hdr = hdu.header
        hdr.update('NUMXVER', __version__, 'Numina package version')
        hdr.update('NUMRNAM', self.__class__.__name__, 'Numina recipe name')
        hdr.update('NUMRVER', self.__version__, 'Numina recipe version')
        
        hdr.update('FILENAME', 'result.fits')
        hdr.update('IMGTYP', 'TARGET', 'Image type')
        hdr.update('NUMTYP', 'TARGET', 'Data product type')
        
        varhdu = pyfits.ImageHDU(sf_data[1], name='VARIANCE')
        num = pyfits.ImageHDU(sf_data[2], name='MAP')

        result = pyfits.HDUList([hdu, varhdu, num])        
        
        _logger.info("Final frame created")
        recipe_result['products'] = [DataFrame(result), SourcesCatalog()]
        
        return recipe_result
    
    def compute_simple_sky(self, targetframes, skyframes, 
                            maxsep=5, step=0, save=True):
        
        # build kdtree        
        sarray = numpy.array([frame.mjd for frame in skyframes])
        # shape must be (n, 1)
        sarray = numpy.expand_dims(sarray, axis=1)
        
        # query
        tarray = numpy.array([frame.mjd for frame in targetframes])
        # shape must be (n, 1)
        tarray = numpy.expand_dims(tarray, axis=1)
        
        kdtree = KDTree(sarray)
        
        # 1 / minutes in a day 
        MIN_TO_DAY = 0.000694444
        _dis, idxs = kdtree.query(tarray, k=1, 
                                 distance_upper_bound=maxsep * MIN_TO_DAY)
        
        for tid, idss in enumerate(idxs):
            try:
                tf = targetframes[tid]
                sf = skyframes[idss]
                self.compute_simple_sky_for_frame(tf, sf, step=step, save=save)
            except IndexError:
                _logger.error('No sky image available for frame %s', tf.lastname)
                raise


    def compute_simple_sky_for_frame(self, frame, skyframe, step=0, save=True):
        _logger.info('Correcting sky in frame %s', frame.lastname)
        _logger.info('with sky computed from frame %s', skyframe.lastname)
        
        if hasattr(skyframe, 'median_sky'):
                sky = skyframe.median_sky
        else:
        
            with pyfits.open(skyframe.lastname, mode='readonly') as hdulist:
                data = hdulist['primary'].data
                valid = data[frame.valid_region]


                if skyframe.objmask_data is not None:
                    _logger.debug('object mask defined')
                    msk = frame.objmask_data
                    sky = numpy.median(valid[msk == 0])
                else:
                    _logger.debug('object mask empty')
                    sky = numpy.median(valid)

            _logger.debug('median sky value is %f', sky)
            skyframe.median_sky = sky
                
        dst = name_skysub_proc(frame.baselabel, step)
        prev = frame.lastname
        
        if save:
            shutil.copyfile(prev, dst)
        else:
            os.rename(prev, dst)
        
        frame.lastname = dst
        
        with pyfits.open(frame.lastname, mode='update') as hdulist:
            data = hdulist['primary'].data
            valid = data[frame.valid_region]
            valid -= sky
        
    def compute_advanced_sky(self, targetframes, objmask, 
                                        skyframes=None, target_is_sky=False,
                                        maxsep=5.0,
                                        nframes=10,
                                        step=0, save=True):
        
        if target_is_sky:
            skyframes = targetframes
            # Each frame is its closets sky frame 
            nframes += 1
        elif skyframes is None:
            raise ValueError('skyframes not defined')
        
        # build kdtree        
        sarray = numpy.array([frame.mjd for frame in skyframes])
        # shape must be (n, 1)
        sarray = numpy.expand_dims(sarray, axis=1)
        
        # query
        tarray = numpy.array([frame.mjd for frame in targetframes])
        # shape must be (n, 1)
        tarray = numpy.expand_dims(tarray, axis=1)
        
        kdtree = KDTree(sarray)
        
        # 1 / minutes in a day 
        MIN_TO_DAY = 0.000694444
        #max_time_sep = self.parameters['sky_images_sep_time'] / 1440.0
        _dis, idxs = kdtree.query(tarray, k=nframes, 
                                 distance_upper_bound=maxsep * MIN_TO_DAY)
        
        nsky = len(sarray)
        
        for tid, idss in enumerate(idxs):
            try:
                tf = targetframes[tid]
                _logger.info('Step %d, SC: computing advanced sky for %s', step, tf.baselabel)
                #filter(lambda x: x < nsky, idss)
                locskyframes = []
                for si in idss:
                    if tid == si:
                        # this sky frame its the current frame, reject
                        continue 
                    if si < nsky:
                        _logger.debug('Step %d, SC: %s is a sky frame', step, skyframes[si].baselabel)
                        locskyframes.append(skyframes[si])
                self.compute_advanced_sky_for_frame(tf, locskyframes, step=step, save=save)
            except IndexError:
                _logger.error('No sky image available for frame %s', tf.lastname)
                raise

    def compute_advanced_sky_for_frame(self, frame, skyframes, step=0, save=True):
        _logger.info('Correcting sky in frame %s', frame.lastname)
        _logger.info('with sky computed from frames')
        for i in skyframes:
            _logger.info('%s', i.flat_corrected)
            
        data = []
        scales = []
        masks = []
        # handle to the FITS file to close it finally
        desc = []
        try:
            for i in skyframes:
                filename = i.flat_corrected
                hdulist = pyfits.open(filename, mode='readonly', memmap=True)
                
                data.append(hdulist['primary'].data[i.valid_region])
                desc.append(hdulist)
                scales.append(numpy.median(data[-1]))
                if i.objmask_data is not None:
                    masks.append(i.objmask_data)
                    _logger.debug('object mask is shared')
                else:
                    maskfile = i.objmask
                    hdulistmask = pyfits.open(maskfile, mode='readonly', memmap=True)
                    masks.append(hdulistmask['primary'].data)
                    desc.append(hdulistmask)
                    _logger.debug('object mask is particular')
                     
            _logger.debug('computing background with %d frames', len(data))
            sky, _, num = median(data, masks, scales=scales)
                
        finally:
            # Closing all FITS files
            for hdl in desc:
                hdl.close()            
            
        if numpy.any(num == 0):
            # We have pixels without
            # sky background information
            _logger.warn('pixels without sky information when correcting %s', frame.flat_corrected)
            binmask = num == 0
            # FIXME: during development, this is faster
            #sky[binmask] = sky[num != 0].mean()
        
            # To continue we interpolate over the patches
            fixpix2(sky, binmask, out=sky, iterations=1)
            name = name_skybackground(frame.baselabel, step)
            pyfits.writeto(name, sky, clobber=True)
            name = name_skybackgroundmask(frame.baselabel, step)
            pyfits.writeto(name, binmask.astype('int16'), clobber=True)        
                
        dst = name_skysub_proc(frame.baselabel, step)
        prev = frame.lastname        
        shutil.copyfile(prev, dst)        
        frame.lastname = dst
        
        with pyfits.open(frame.lastname, mode='update') as hdulist:
            data = hdulist['primary'].data
            valid = data[frame.valid_region]
            valid -= sky



    def combine_frames(self, frames, out=None, step=0):
        _logger.debug('Step %d, opening sky-subtracted frames', step)

        def fits_open(name):
            '''Open FITS with memmap in readonly mode'''
            return pyfits.open(name, mode='readonly', memmap=True)

        frameslll = [fits_open(frame.lastname) for frame in frames if frame.valid_target]
        _logger.debug('Step %d, opening mask frames', step)
        mskslll = [fits_open(frame.resized_mask) for frame in frames if frame.valid_target]
        _logger.debug('Step %d, combining %d frames', step, len(frameslll))
        try:
            extinc = [pow(10, -0.4 * frame.airmass * self.parameters['extinction']) for frame in frames if frame.valid_target]
            data = [i['primary'].data for i in frameslll]
            masks = [i['primary'].data for i in mskslll]
            
            out = quantileclip(data, masks, scales=extinc, dtype='float32', out=out, fclip=0.1)
            
            # saving the three extensions
            pyfits.writeto('result_i%0d.fits' % step, out[0], clobber=True)
            pyfits.writeto('result_var_i%0d.fits' % step, out[1], clobber=True)
            pyfits.writeto('result_npix_i%0d.fits' % step, out[2], clobber=True)
                
            return out
            
        finally:
            _logger.debug('Step %d, closing sky-subtracted frames', step)
            map(lambda x: x.close(), frameslll)
            _logger.debug('Step %d, closing mask frames', step)
            map(lambda x: x.close(), mskslll)
            
    def apply_superflat(self, frames, flatdata, step=0, save=True):
        _logger.info("Step %d, SF: apply superflat", step)

        # Process all frames with the fitted flat
        # FIXME: not sure
        for frame in frames:
            self.correct_superflat(frame, flatdata, step=step, save=save)
        return frames
            
    def correct_superflat(self, frame, fitted, step=0, save=True):
        
        frame.flat_corrected = name_skyflat_proc(frame.baselabel, step)
        
        if save:
            shutil.copyfile(frame.resized_base, frame.flat_corrected)
        else:
            os.rename(frame.resized_base, frame.flat_corrected)
        
        _logger.info("Step %d, SF: apply superflat to frame %s", step, frame.flat_corrected)
        with pyfits.open(frame.flat_corrected, mode='update') as hdulist:
            data = hdulist['primary'].data
            datar = data[frame.valid_region]
            data[frame.valid_region] = correct_flatfield(datar, fitted)

            frame.lastname = frame.flat_corrected            
            
            # FIXME: plotting
            try:
                self.figure_image(data[frame.valid_region], frame)
            except ValueError:
                _logger.warning('Problem plotting %s', frame.lastname)
                        
    def compute_superflat(self, frames, channels, target_is_sky, segmask=None, step=0):
        _logger.info("Step %d, SF: combining the frames without offsets", step)
        try:
            filelist = []
            data = []
            for frame in frames:
                _logger.debug('Step %d, opening resized frame %s', step, frame.resized_base)
                hdulist = pyfits.open(frame.resized_base, memmap=True, mode='readonly')
                filelist.append(hdulist)
                data.append(hdulist['primary'].data[frame.valid_region])

            scales = [frame.median_scale for frame in frames]
            
            # FIXME: plotting
            self.figure_median_background(scales)
            
            masks = None
            if segmask is not None:
                if target_is_sky:
                    masks = [segmask[frame.valid_region] for frame in frames]
                else:
                    masks = [pyfits.getdata(frame.objmask) for frame in frames]
            _logger.debug('Step %d, combining %d frames', step, len(data))
            sf_data, _sf_var, sf_num = flatcombine(data, masks, scales=scales, 
                                                    blank=1.0 / scales[0])            
        finally:
            _logger.debug('Step %d, closing resized frames and mask', step)
            for fileh in filelist:               
                fileh.close()            

        # We interpolate holes by channel
        _logger.debug('Step %d, interpolating holes by channel', step)
        for channel in channels:
            mask = (sf_num[channel] == 0)
            if numpy.any(mask):                    
                fixpix2(sf_data[channel], mask, out=sf_data[channel])

        # Normalize, flat has mean = 1
        sf_data /= sf_data.mean()
        
        # Auxiliary data
        sfhdu = pyfits.PrimaryHDU(sf_data)            
        sfhdu.writeto(name_skyflat('comb', step), clobber=True)
        return sf_data
        
    def update_scale_factors(self, frames, step=0):
        _logger.info('Step %d, SF: computing scale factors', step)
        # FIXME: not sure
        for frame in frames:
            region = frame.valid_region
            data = pyfits.getdata(frame.resized_base)[region]
            mask = pyfits.getdata(frame.resized_mask)[region]
            # FIXME: while developing this ::10 is faster, remove later            
            frame.median_scale = numpy.median(data[mask == 0][::10])
            _logger.debug('median value of %s is %f', frame.resized_base, frame.median_scale)
        return frames
        
    def resize(self, frames, shape, offsetsp, finalshape, window=None, scale=1, step=0):
        _logger.info('Resizing frames and masks')
        for frame, rel_offset in zip(frames, offsetsp):
            if frame.valid_target:
                region, _ = subarray_match(finalshape, rel_offset, shape)
                # Valid region
                frame.valid_region = region
                # Relative offset
                frame.rel_offset = rel_offset
                # names of frame and mask
                framen, maskn = name_redimensioned_frames(frame.baselabel, step)
                frame.resized_base = framen
                frame.resized_mask = maskn
                _logger.debug('%s, valid region is %s, relative offset is %s', frame.label, 
                custom_region_to_str(region), rel_offset)
                self.resize_frame_and_mask(frame, finalshape, framen, maskn, window, scale)
                
        return frames

    def resize_frame_and_mask(self, frame, finalshape, framen, maskn, window, scale):
        _logger.info('Resizing frame %s, window=%s, subpix=%i', frame.label, 
                     custom_region_to_str(window), scale)
        resize_fits(frame.label, framen, finalshape, frame.valid_region, 
                    window=window, scale=scale)

        _logger.info('Resizing mask %s, subpix x%i', frame.label, scale)
        # We don't conserve the sum of the values of the frame here, just
        # expand the mask
        resize_fits(frame.mask, maskn, finalshape, frame.valid_region, 
                    fill=1, window=window, scale=scale, conserve=False)
        
    def figure_init(self, shape):
        self._figure.clf()
        ax = self._figure.add_subplot(111)
        cmap = mpl.cm.get_cmap('gray')
        norm = mpl.colors.LogNorm()
        ax.imshow(numpy.ones(shape), cmap=cmap, norm=norm)
        ax.set_xlabel('X')
        ax.set_ylabel('Y')
        self._figure.canvas.draw()
        
    def figures_after_combine(self, sf_data, step=0):
       
        # FIXME, more plots
        def truncated(array, frac=0.1):
            '''Dirty truncated mean'''
            nf = int(array.size * frac)
            array.sort()
            new = array[nf:-nf]
            lnew = len(new)
            if lnew == 0:
                return 0, 0
            elif lnew == 1:
                return new.mean(), 0.0
            else:
                return new.mean(), new.std()
            
        ndata = sf_data[2].astype('int')                    
        data = sf_data[0]
        
        nimages = ndata.max()

        rnimage = range(1, nimages + 1)
        rmean = rnimage[:]
        rstd = rnimage[:]
            
        for pix in rnimage:
            rmean[pix - 1], rstd[pix - 1] = truncated(data[ndata == pix])                
            
        avg_rms = self.figure_check_combination(rnimage, rmean, rstd, step=step)
                        
        # Fake sky error image
        self.figure_simple_image(ndata, title='Number of frames combined')

        # Create fake error frame
        mask = (ndata <= 0)
        ndata[mask] = 1
        fake = numpy.where(mask, 0.0, numpy.random.normal(avg_rms / numpy.sqrt(ndata)))
        ndata[mask] = 0
        self.figure_simple_image(fake, title='Fake sky error image')
        # store fake image
        pyfits.writeto('fake_sky_rms_i%0d.fits' % step, fake, clobber=True)
        
    def figure_check_combination(self, rnimage, rmean, rstd, step=0):            
        self._figure.clf()
        self._figure.subplots_adjust(hspace=0.001)
        
        ax1 = self._figure.add_subplot(3,1,1)
        pred = [rstd[-1] * math.sqrt(rnimage[-1] / float(npix)) for npix in rnimage]
        ax1.plot(rnimage, rstd, 'g*', rnimage, pred, 'y-')
        ax1.set_title("")
        ax1.set_ylabel('measured sky rms')
        
        ax2 = self._figure.add_subplot(3,1,2, sharex=ax1)
        pred = [val * math.sqrt(npix) for val, npix in zip(rstd, rnimage)]
        avg_rms = sum(pred) / len(pred)
        ax2.plot(rnimage, pred, 'r*', [rnimage[0], rnimage[-1]], [avg_rms,avg_rms])
        ax2.set_ylabel('scaled sky rms')

        ax3 = self._figure.add_subplot(3,1,3, sharex=ax1)
        ax3.plot(rnimage, rmean, 'b*')
        ax3.set_ylabel('mean sky')
        ax3.set_xlabel('number of frames per pixel')

        xticklabels = ax1.get_xticklabels() + ax2.get_xticklabels()
        mpl.artist.setp(xticklabels, visible=False)
        self._figure.canvas.draw()
        self._figure.savefig('figure-check-combination_i%01d.png' % step)
        return avg_rms        
        
    def figure_simple_image(self, data, title=None):
        self._figure.clf()
        ax = self._figure.add_subplot(111)
        cmap = mpl.cm.get_cmap('gray')
        norm = mpl.colors.LogNorm()
        if title is not None:
            ax.set_title(title)
                          
        ax.set_xlabel('X')
        ax.set_ylabel('Y')            
        ax.imshow(data, cmap=cmap, norm=norm)                                
        self._figure.canvas.draw()
  
    def figure_median_background(self, scales, step=0):
        # FIXME: plotting
        self._figure.clf()
        ax = self._figure.add_subplot(1,1,1) 
        ax.plot(scales, 'r*')
        ax.set_title("")
        ax.set_xlabel('Image number')
        ax.set_ylabel('Median')
        self._figure.canvas.draw()
        self._figure.savefig('figure-median-sky-background_i%01d.png' % step)
        
    def figure_image(self, thedata, image):
        # FIXME: remove this dependency
        import numdisplay        
        ax = self._figure.gca()
        image_axes, = ax.get_images()
        image_axes.set_data(thedata)
        z1, z2 = numdisplay.zscale.zscale(thedata)
        image_axes.set_clim(z1, z2)
        clim = image_axes.get_clim()
        ax.set_title('%s, bg=%g fg=%g, linscale' % (image.lastname, clim[0], clim[1]))        
        self._figure.canvas.draw()      

    def create_mask_single(self, frame, seeing_fwhm, step=0):
        #
        #remove_border = True
        
        # sextractor takes care of bad pixels
        sex = SExtractor()
        sex.config['CHECKIMAGE_TYPE'] = "SEGMENTATION"
        sex.config["CHECKIMAGE_NAME"] = name_object_mask(frame.baselabel, step) 
        

        sex.config['VERBOSE_TYPE'] = 'QUIET'
        sex.config['PIXEL_SCALE'] = 1
        sex.config['BACK_TYPE'] = 'AUTO' 

        if seeing_fwhm is not None and seeing_fwhm > 0:
            sex.config['SEEING_FWHM'] = seeing_fwhm * sex.config['PIXEL_SCALE']

        sex.config['PARAMETERS_LIST'].append('FLUX_BEST')
        sex.config['PARAMETERS_LIST'].append('X_IMAGE')
        sex.config['PARAMETERS_LIST'].append('Y_IMAGE')
        sex.config['PARAMETERS_LIST'].append('A_IMAGE')
        sex.config['PARAMETERS_LIST'].append('B_IMAGE')
        sex.config['PARAMETERS_LIST'].append('THETA_IMAGE')
        sex.config['PARAMETERS_LIST'].append('FWHM_IMAGE')
        sex.config['PARAMETERS_LIST'].append('CLASS_STAR')                
        
        filename = frame.lastname
        
        # Lauch SExtractor on a FITS file
        sex.run(filename)
        
        # Plot objects
        # FIXME, plot sextractor objects on top of image
        patches = []
        fwhms = []
        nfirst = 0
        catalog_f = sopen(sex.config['CATALOG_NAME'])
        try:
            star = catalog_f.readline()
            while star:
                flags = star['FLAGS']
                # ignoring those objects with corrupted apertures
                if flags & sexcatalog.CORRUPTED_APER:
                    star = catalog_f.readline()
                    continue
                center = (star['X_IMAGE'], star['Y_IMAGE'])
                wd = 10 * star['A_IMAGE']
                hd = 10 * star['B_IMAGE']
                color = 'red'
                e = Ellipse(center, wd, hd, star['THETA_IMAGE'], color=color)
                patches.append(e)
                fwhms.append(star['FWHM_IMAGE'])
                nfirst += 1
                # FIXME Plot a ellipse
                star = catalog_f.readline()
        finally:
            catalog_f.close()
            
        p = PatchCollection(patches, alpha=0.4)
        ax = self._figure.gca()
        ax.add_collection(p)
        self._figure.canvas.draw()
        self._figure.savefig('figure-sky-segmentation-overlay_%01d.png' % step)

        self.figure_fwhm_histogram(fwhms, step=step)
                    
        # mode with an histogram
        hist, edges = numpy.histogram(fwhms, 50)
        idx = hist.argmax()
        
        seeing_fwhm = 0.5 * (edges[idx] + edges[idx + 1]) 
        if seeing_fwhm <= 0:
            _logger.warning('Seeing FHWM %f pixels is negative, reseting', seeing_fwhm)
            seeing_fwhm = None
        else:
            _logger.info('Seeing FHWM %f pixels (%f arcseconds)', seeing_fwhm, seeing_fwhm * sex.config['PIXEL_SCALE'])
        name_segmask(step)
        _logger.info('Step %d, create object mask %s', step,  frame.objmask)
        frame.objmask = name_object_mask(frame.baselabel, step)                 
        frame.objmask_data = None
        return frame, seeing_fwhm
        
    def create_mask(self, sf_data, seeing_fwhm, step=0):
        # FIXME more plots
        self.figure_final_before_s(sf_data[0])

        #
        remove_border = True
        
        # sextractor takes care of bad pixels
        sex = SExtractor()
        sex.config['CHECKIMAGE_TYPE'] = "SEGMENTATION"
        sex.config["CHECKIMAGE_NAME"] = name_segmask(step)
        sex.config['VERBOSE_TYPE'] = 'QUIET'
        sex.config['PIXEL_SCALE'] = 1
        sex.config['BACK_TYPE'] = 'AUTO' 

        if seeing_fwhm is not None and seeing_fwhm > 0:
            sex.config['SEEING_FWHM'] = seeing_fwhm * sex.config['PIXEL_SCALE']

        sex.config['PARAMETERS_LIST'].append('FLUX_BEST')
        sex.config['PARAMETERS_LIST'].append('X_IMAGE')
        sex.config['PARAMETERS_LIST'].append('Y_IMAGE')
        sex.config['PARAMETERS_LIST'].append('A_IMAGE')
        sex.config['PARAMETERS_LIST'].append('B_IMAGE')
        sex.config['PARAMETERS_LIST'].append('THETA_IMAGE')
        sex.config['PARAMETERS_LIST'].append('FWHM_IMAGE')
        sex.config['PARAMETERS_LIST'].append('CLASS_STAR')                
        if remove_border:
            weigthmap = 'weights4rms.fits'
            
            # Create weight map, remove n pixs from either side
            # using a Hannig filter
            # npix = 90          
            # w1 = npix
            # w2 = npix
            # wmap = numpy.ones_like(sf_data[0])
            
            # cos_win1 = numpy.hanning(2 * w1)
            # cos_win2 = numpy.hanning(2 * w2)
                                   
            # wmap[:,:w1] *= cos_win1[:w1]                    
            # wmap[:,-w1:] *= cos_win1[-w1:]
            # wmap[:w2,:] *= cos_win2[:w2, numpy.newaxis]
            # wmap[-w2:,:] *= cos_win2[-w2:, numpy.newaxis]                 
            
            # Take the number of combined images from the combined image
            wm = sf_data[2].copy()
            # Dont search objects where nimages < lower
            # FIXME: this is a magic number
            # We ignore objects in regions where we have less
            # than 10% of the images
            lower = sf_data[2].max() / 10
            wm[wm < lower] = 0
            pyfits.writeto(weigthmap, wm, clobber=True)
                                
            sex.config['WEIGHT_TYPE'] = 'MAP_WEIGHT'
            # FIXME: this is a magic number
            # sex.config['WEIGHT_THRESH'] = 50
            sex.config['WEIGHT_IMAGE'] = weigthmap
        
        filename = 'result_i%0d.fits' % (step)
        
        # Lauch SExtractor on a FITS file
        sex.run(filename)
        
        # Plot objects
        # FIXME, plot sextractor objects on top of image
        patches = []
        fwhms = []
        nfirst = 0
        catalog_f = sopen(sex.config['CATALOG_NAME'])
        try:
            star = catalog_f.readline()
            while star:
                flags = star['FLAGS']
                # ignoring those objects with corrupted apertures
                if flags & sexcatalog.CORRUPTED_APER:
                    star = catalog_f.readline()
                    continue
                center = (star['X_IMAGE'], star['Y_IMAGE'])
                wd = 10 * star['A_IMAGE']
                hd = 10 * star['B_IMAGE']
                color = 'red'
                e = Ellipse(center, wd, hd, star['THETA_IMAGE'], color=color)
                patches.append(e)
                fwhms.append(star['FWHM_IMAGE'])
                nfirst += 1
                # FIXME Plot a ellipse
                star = catalog_f.readline()
        finally:
            catalog_f.close()
            
        p = PatchCollection(patches, alpha=0.4)
        ax = self._figure.gca()
        ax.add_collection(p)
        self._figure.canvas.draw()
        self._figure.savefig('figure-segmentation-overlay_%01d.png' % step)

        self.figure_fwhm_histogram(fwhms, step=step)
                    
        # mode with an histogram
        hist, edges = numpy.histogram(fwhms, 50)
        idx = hist.argmax()
        
        seeing_fwhm = 0.5 * (edges[idx] + edges[idx + 1]) 
        if seeing_fwhm <= 0:
            _logger.warning('Seeing FHWM %f pixels is negative, reseting', seeing_fwhm)
            seeing_fwhm = None
        else:
            _logger.info('Seeing FHWM %f pixels (%f arcseconds)', seeing_fwhm, seeing_fwhm * sex.config['PIXEL_SCALE'])
        objmask = pyfits.getdata(name_segmask(step))
        return objmask, seeing_fwhm
    
    def figure_final_before_s(self, data):
        import numdisplay
        self._figure.clf()
        ax = self._figure.add_subplot(111)
        cmap = mpl.cm.get_cmap('gray')
        #norm = mpl.colors.LogNorm()
        ax.set_title('Result image')              
        ax.set_xlabel('X')
        ax.set_ylabel('Y')
        z1, z2 = numdisplay.zscale.zscale(data)
        ax.imshow(data, cmap=cmap, clim=(z1, z2))                                
        self._figure.canvas.draw()
    
    def figure_fwhm_histogram(self, fwhms, step=0):
        self._figure.clf()
        ax = self._figure.add_subplot(111)
        ax.set_title('FWHM of objects')
        ax.hist(fwhms, 50, normed=1, facecolor='g', alpha=0.75)
        self._figure.canvas.draw()
        self._figure.savefig('figure-fwhm-histogram_i%01d.png' % step)

        
        



#
# Copyright 2008-2012 Universidad Complutense de Madrid
# 
# This file is part of PyEmir
# 
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
# 

import itertools as ito

import numpy # pylint: disable-msgs=E1101

from numina.extraiter import braid
from numina.treedict import TreeDict
from numina.instrument.detector import nIRDetector, Amplifier, Das
from numina.instrument.detector import SingleReadoutMode
from numina.instrument.detector import CdsReadoutMode
from numina.instrument.detector import RampReadoutMode
from numina.instrument.detector import FowlerReadoutMode

def _channel_gen1(beg, end, step):
    return ito.imap(lambda x: (x, x + step), xrange(beg, end, step))

def _channel_gen2(beg, end, step):
    return ito.imap(lambda x: (x - step, x), xrange(beg, end, -step))

def _ch1():
    return ito.izip(ito.repeat(slice(1024, 2048)), ito.starmap(slice, _channel_gen2(1024, 0, 128)))

def _ch2():  
    return ito.izip(ito.starmap(slice, _channel_gen2(1024, 0, 128)), ito.repeat(slice(0, 1024)))

def _ch3():
    return ito.izip(ito.repeat(slice(0, 1024)), ito.starmap(slice, _channel_gen1(1024, 2048, 128)))

def _ch4():
    return ito.izip(ito.starmap(slice, _channel_gen1(1024, 2048, 128)), ito.repeat(slice(1024, 2048)))

# Channels are listed per quadrant and then in fast readout order
CHANNELS = list(ito.chain(_ch1(), _ch2(), _ch3(), _ch4()))

# Channels in read out order
CHANNELS_READOUT = list(braid(_ch1(), _ch2(), _ch3(), _ch4()))

# Quadrants are listed starting at left-top and counter-clockwise then
QUADRANTS = [(slice(1024, 2048), slice(0, 1024)),
             (slice(0, 1024), slice(0, 1024)),
             (slice(0, 1024), slice(1024, 2048)),
             (slice(1024, 2048), slice(1024, 2048))
             ]

class EmirDas(Das):
    def __init__(self, detector):
        Das.__init__(self, detector)
                
    def readout_mode_single(self, repeat=1):
        self.readmode(SingleReadoutMode(repeat=repeat)) 
    
    def readout_mode_cds(self, repeat=1):
        mode = CdsReadoutMode(repeat=repeat)
        self.readmode(mode)
    
    def readout_mode_fowler(self, reads, repeat=1, readout_time=0.0):
        mode = FowlerReadoutMode(reads, repeat=repeat, 
                                 readout_time=readout_time)
        self.readmode(mode)    

    def readout_mode_ramp(self, reads, repeat=1):
        mode = RampReadoutMode(reads, repeat=repeat)
        self.readmode(mode)

class Hawaii2Detector(nIRDetector):
    '''Hawaii2 detector.'''
    
    AMP1 = QUADRANTS # 1 amplifier per quadrant
    AMP8 = CHANNELS # 8 amplifiers per quadrant
    
    shape = (2048, 2048)
    amplifiers = CHANNELS
    
    def __init__(self, gain=1.0, ron=0.0, dark=1.0, well=65535,
                 pedestal=200., flat=1.0, resetval=0, resetnoise=0.0,
                 ampmode='8'):
        '''
            :parameter gain: gain in e-/ADU
            :parameter ron: ron in ADU
            :parameter dark: dark current in e-/s
            :parameter well: well depth in ADUs 
        '''
        
        
        if ampmode not in ['1', '8']:
            raise ValueError('ampmode must be "1" or "8"')
        
        self.ampmode = ampmode
        # Amplifier region
        self.ampgeom = self.AMP1 if ampmode == '1' else self.AMP8
        
        ampgain = ito.cycle(numpy.asarray(gain).flat)
        ampron = ito.cycle(numpy.asarray(ron).flat)
        ampwell = ito.cycle(numpy.asarray(ron).flat)
        amps = [Amplifier(geom, gain, ron, well) for geom, gain, ron, well in zip(self.amplifiers, ampgain, ampron, ampwell)]
        
        nIRDetector.__init__(self, self.shape, amps, dark, pedestal, flat, resetval, resetnoise)
        
class EmirDetector(Hawaii2Detector):
    def __init__(self, flat=1.0):        
        # ADU, per AMP
        ron = [3.14, 3.060, 3.09, 3.08, 3.06, 3.20, 3.13, 3.10, 2.98, 2.98, 
               2.95, 3.06, 3.050, 3.01, 3.05, 3.20, 2.96, 3.0, 3.0, 2.99, 3.14, 
               2.99, 3.06, 3.05, 3.08, 3.06, 3.01, 3.02, 3.07, 2.99, 3.03, 3.06]

        # e-/ADU per AMP
        gain = [2.83, 2.84, 2.82, 2.92, 2.98, 2.71, 3.03, 2.92, 3.04, 2.79, 2.72, 
                3.07, 3.03, 2.91, 3.16, 3.22, 2.93, 2.88, 2.99, 3.24, 3.25, 3.12, 
                3.29, 3.03, 3.04, 3.35, 3.11, 3.25, 3.29, 3.17, 2.93, 3.05]

        # e-/s global median
        dark = 0.298897832632

        # ADU per AMP
        wdepth = [42353.1, 42148.3, 42125.5, 42057.9, 41914.1, 42080.2, 42350.3, 
                  41830.3, 41905.3, 42027.9, 41589.5, 41712.7, 41404.9, 41068.5, 
                  40384.9, 40128.1, 41401.4, 41696.5, 41461.1, 41233.2, 41351.0, 
                  41803.7, 41450.2, 41306.2, 41609.4, 41414.1, 41324.5, 41691.1, 
                  41360.0, 41551.2, 41618.6, 41553.5]
        
        self.meta = TreeDict()
        self.meta['gain'] = 2.8
        self.meta['readnoise'] = 3.0

        Hawaii2Detector.__init__(self, gain=gain, ron=ron, dark=dark, 
                                           well=wdepth, flat=flat)

if __name__ == '__main__':
    
    
    det = EmirDetector()



#
# Copyright 2008-2012 Universidad Complutense de Madrid
# 
# This file is part of PyEmir
# 
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

'''
    Recipes for finding the best focus.

'''

import logging

from numina.recipes import RecipeBase, Parameter, provides
from numina.recipes import DataProductParameter
from numina.logger import log_to_history

from ..dataproducts import MasterBias, MasterDark, MasterBadPixelMask
from ..dataproducts import TelescopeFocus
from ..dataproducts import DTUFocus
from ..dataproducts import MasterIntensityFlat
from ..dataproducts import NonLinearityCalibration

__all__ = ['TelescopeRoughFocusRecipe', 
           'TelescopeFineFocusRecipe',
           'DTUFocusRecipe',           
           ]

_logger = logging.getLogger('emir.recipes')

@provides(TelescopeFocus)
class TelescopeRoughFocusRecipe(RecipeBase):
    '''Recipe to compute the telescope focus.
    
    **Observing modes:**

     * Telescope rough focus
     * Emir focus control 

    **Inputs:**

     * A list of images
     * A list of sky images
     * Bias, dark, flat
     * A model of the detector
     * List of focii 

    **Outputs:**
     * Best focus
    '''
    __requires__ = [       
        DataProductParameter('master_bias', MasterBias, 'Master bias image'),
        DataProductParameter('master_dark', MasterDark, 'Master dark image'),
        DataProductParameter('master_bpm', MasterBadPixelMask, 'Master bad pixel mask'),
        DataProductParameter('nonlinearity', NonLinearityCalibration([1.0, 0.0]), 
                  'Polynomial for non-linearity correction'),
        DataProductParameter('master_intensity_ff', MasterIntensityFlat, 
                  'Master intensity flatfield'),
        Parameter('objects', None, 'List of x-y pair of object coordinates'),
        Parameter('focus_range', None, 'Focus range: begin, end and step')        
    ]

    def __init__(self):
        super(TelescopeRoughFocusRecipe, self).__init__(
            author="Sergio Pascual <sergiopr@fis.ucm.es>",
            version="0.1.0"
        )

    def run(self, obresult):
        return {'products': [TelescopeFocus()]}

@provides(TelescopeFocus)
class TelescopeFineFocusRecipe(RecipeBase):
    '''
    Recipe to compute the telescope focus.
    
    **Observing modes:**

        * Telescope fine focus 

    **Inputs:**

     * A list of images
     * A list of sky images
     * Bias, dark, flat
     * A model of the detector
     * List of focii 

    **Outputs:**
     * Best focus
    
    '''

    __requires__ = [       
        DataProductParameter('master_bias', MasterBias, 'Master bias image'),
        DataProductParameter('master_dark', MasterDark, 'Master dark image'),
        DataProductParameter('master_bpm', MasterBadPixelMask, 'Master bad pixel mask'),
        DataProductParameter('nonlinearity', NonLinearityCalibration([1.0, 0.0]), 
                  'Polynomial for non-linearity correction'),
        DataProductParameter('master_intensity_ff', MasterIntensityFlat, 
                  'Master intensity flatfield'),
        Parameter('objects', None, 'List of x-y pair of object coordinates'),        
    ]

    def __init__(self):
        super(TelescopeFineFocusRecipe, self).__init__(
            author="Sergio Pascual <sergiopr@fis.ucm.es>",
            version="0.1.0"
        )

    def run(self, obresult):
        return {'products': [TelescopeFocus()]}

@provides(DTUFocus)
class DTUFocusRecipe(RecipeBase):
    '''
    Recipe to compute the DTU focus.
    
    **Observing modes:**

        * EMIR focus control

    **Inputs:**

     * A list of images
     * A list of sky images
     * Bias, dark, flat
     * A model of the detector
     * List of focii 

    **Outputs:**
     * Best focus
    
    '''

    __requires__ = [       
        DataProductParameter('master_bias', MasterBias, 'Master bias image'),
        DataProductParameter('master_dark', MasterDark, 'Master dark image'),
        DataProductParameter('master_bpm', MasterBadPixelMask, 'Master bad pixel mask'),
        DataProductParameter('nonlinearity', NonLinearityCalibration([1.0, 0.0]), 
                  'Polynomial for non-linearity correction'),
        DataProductParameter('master_intensity_ff', MasterIntensityFlat, 
                  'Master intensity flatfield'),
        Parameter('objects', None, 'List of x-y pair of object coordinates'),
        Parameter('msm_pattern', None, 'List of x-y pair of slit coordinates'),
        Parameter('dtu_focus_range', None, 
                  'Focus range of the DTU: begin, end and step')        
    ]

    def __init__(self):
        super(DTUFocusRecipe, self).__init__(
            author="Sergio Pascual <sergiopr@fis.ucm.es>",
            version="0.1.0"
        )

    def run(self, obresult):
        return {'products': [DTUFocus()]}



#
# Copyright 2008-2012 Universidad Complutense de Madrid
# 
# This file is part of PyEmir
# 
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
# 

'''Data products produced by the EMIR pipeline.'''

import logging
import pyfits

from numina.recipes import DataFrame, DataProduct

from .simulator import EmirImageFactory

_logger = logging.getLogger('emir.dataproducts')

class MasterBadPixelMask(DataFrame):
    def __init__(self, hdu):
        super(MasterBias, self).__init__(hdu)

    def metadata(self):
        hdr = self.image[0].header
        yield 'detector0.mode', hdr['ccdmode']

class MasterBias(DataFrame):
    '''Master bias product
    
    This image has 4 extensions: primary, two variance extensions
    and number of pixels used in the combination.
    
    The variance extensions are computed using two different methods. 
    The first one is the variance of the same pixels in different images.
    The second extension is the variance of each channel in the final image.
    
    
    '''
    def __init__(self, hdu):
        super(MasterBias, self).__init__(hdu)

    def metadata(self):
        hdr = self.image[0].header
        yield 'detector0.mode', hdr['ccdmode']

class MasterDark(DataFrame):
    '''Master dark product
    
    This image has 4 extensions: primary, two variance extensions
    and number of pixels used in the combination.
    
    The variance extensions are computed using two different methods. 
    The first one is the variance of the same pixels in different images.
    The second extension is the variance of each channel in the final image.
    
    
    '''
    def __init__(self, hdu):
        super(MasterDark, self).__init__(hdu)

    def metadata(self):
        hdr = self.image[0].header
        yield 'detector0.mode', hdr['ccdmode']

class MasterIntensityFlat(DataFrame):
    def __init__(self, hdu):
        super(MasterIntensityFlat, self).__init__(hdu)

    def metadata(self):
        hdr = self.image[0].header
        yield 'detector0.mode', hdr['ccdmode']
        yield 'filter0', hdr['filter']
        
class MasterSpectralFlat(DataFrame):
    def __init__(self, hdu):
        super(MasterSpectralFlat, self).__init__(hdu)

    def metadata(self):
        hdr = self.image[0].header
        yield 'detector0.mode', hdr['ccdmode']
        yield 'filter0', hdr['filter']

class Spectra(DataFrame):
    def __init__(self, hdu):
        super(Spectra, self).__init__(hdu)

    def metadata(self):
        hdr = self.image[0].header
        yield 'detector0.mode', hdr['ccdmode']
        yield 'filter0', hdr['filter']
        
class DataCube(DataFrame):
    def __init__(self, hdu):
        super(DataFrame, self).__init__(None)

class TelescopeFocus(DataProduct):
    pass

class DTUFocus(DataProduct):
    pass

class DTU_XY_Calibration(DataProduct):
    pass

class DTU_Z_Calibration(DataProduct):
    pass

class DTUFlexureCalibration(DataProduct):
    pass

class SlitTransmissionCalibration(DataProduct):
    pass

class WavelengthCalibration(DataProduct):
    pass

class CSU2DetectorCalibration(DataProduct):
    pass

class PointingOriginCalibration(DataProduct):
    pass

class SpectroPhotometricCalibration(DataProduct):
    pass

class PhotometricCalibration(DataProduct):
    pass

class MasterGainMap(DataProduct):
    pass

class MasterRONMap(DataProduct):
    pass

class NonLinearityCalibration(DataProduct):
    def __init__(self, poly):
        super(NonLinearityCalibration, self).__init__()
        self.poly = poly

class TelescopeOffset(DataProduct):
    pass

class MSMPositions(DataProduct):
    pass

class SourcesCatalog(DataProduct):
    pass

class LinesCatalog(DataProduct):
    pass

class ChannelLevelStatistics(DataProduct):
    ''' A list of exposure time, mean, std dev and median per channel'''
    def __init__(self, exposure):
        self.exposure = exposure
        self.statistics = []

# DataFrame -> Raw: PRIMARY
#       -> Result: PRIMARY, VARIANCE, MAP 

_result_types = ['image', 'spectrum']
_extensions = ['primary', 'variance', 'map', 'wcs']

def create_raw(data=None, headers=None, default_headers=None, imgtype='image'):
    
    if imgtype not in _result_types:
        raise TypeError('%s not in %s' % (imgtype, _result_types))
    
    hdefault = default_headers or EmirImageFactory.default
    
    hdu = pyfits.PrimaryHDU(data, hdefault[imgtype]['primary'])
                
    if headers is not None:
        _logger.info('Updating keywords in %s header', 'PRIMARY')      
        for key in headers:
            _logger.debug('Updating keyword %s with value %s', 
                          key, headers[key])
            hdu.header.update(key, headers[key])
    return pyfits.HDUList([hdu])


def create_result(data=None, headers=None, 
                   variance=None, variance_headers=None,
                   exmap=None, exmap_headers=None,
                   default_headers=None, imgtype='image'):
    hdulist = create_raw(data, headers, default_headers, imgtype)
    extensions = {}
    extensions['variance'] = (variance, variance_headers)
    extensions['map'] = (exmap, exmap_headers)
    
    hdefault = default_headers or EmirImageFactory.default
    
    for extname in ['variance', 'map']:
        edata = extensions[extname]
        hdu = pyfits.ImageHDU(edata[0], hdefault[imgtype][extname], name=extname)
        hdulist.append(hdu)
    return hdulist




#
# Copyright 2011-2012 Universidad Complutense de Madrid
# 
# This file is part of PyEmir
# 
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

'''
Mosiac image mode recipe of EMIR

'''

import logging

from numina.recipes import RecipeBase, Parameter, provides, DataFrame

from emir.dataproducts import SourcesCatalog

_logger = logging.getLogger('numina.recipes.emir')

@provides(DataFrame, SourcesCatalog)
class MosaicRecipe(RecipeBase):
    '''
    The effect of recording a series of stare images, with the same
    acquisition parameters, and taken by pointing to a number of
    sky positions, with separations of the order of the EMIR FOV.
    This command is designed to fully cover a given area on the
    sky, but can also be used to point to a number of sky positions
    on which acquisition is only made at the beginning. Supersky
    frame(s) can be built from the image series.

    **Observing modes:**

        * Mosiac images
    
    '''

    __requires__ = [
        # FIXME: this parameter is optional 
        Parameter('sources', None, 
                  'List of x, y coordinates to measure FWHM')
    ]

    def __init__(self):
        super(MosaicRecipe, self).__init__(
            author="Sergio Pascual <sergiopr@fis.ucm.es>",
            version="0.1.0"
        )

    def run(self, obresult):
        return {'products': [DataFrame(None), SourcesCatalog()]}





#
# Copyright 2008-2012 Universidad Complutense de Madrid
# 
# This file is part of PyEmir
# 
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
# 

'''Recipes for EMIR Observing Modes.'''

from numina.pipeline import Pipeline

from .auxiliary import BiasRecipe, DarkRecipe, IntensityFlatRecipe
from .image import StareImageRecipe, DitheredImageRecipe, MicroditheredImageRecipe
from .image import NBImageRecipe, MosaicRecipe

_equiv_class = {
    'bias_image': BiasRecipe,
    'dark_image': DarkRecipe,
    'intensity_flatfield': IntensityFlatRecipe,
    'msm_spectral_flatfield': 'auxiliary:Recipe',
    'slit_transmission_calibration': 'auxiliary:Recipe',
    'wavelength_calibration': 'auxiliary:Recipe',
    'ts_rough_focus': 'auxiliary:Recipe',
    'ts_fine_focus': 'auxiliary:Recipe',
    'emir_focus_control': 'auxiliary:Recipe',
    'image_setup': 'auxiliary:Recipe',
    'mos_and_longslit_setup': 'auxiliary:Recipe',
    'target_acquisition': 'auxiliary:Recipe',
    'mask_imaging': 'auxiliary:Recipe',
    'msm_and_lsm_check': 'auxiliary:Recipe',
    'stare_image': StareImageRecipe,
    'nb_image': NBImageRecipe,
    'dithered_image': DitheredImageRecipe,
    'microdithered_image': MicroditheredImageRecipe,
    'mosaiced_image': MosaicRecipe,
    'stare_spectra': 'mos:Recipe',
    'dn_spectra': 'mos:Recipe',
    'offset_spectra': 'mos:Recipe',
    'raster_spectra': 'ls:Recipe',
}

class EmirPipeline(Pipeline):
    def __init__(self, version):
        super(EmirPipeline, self).__init__(name='emir', 
                version=version,
                recipes=_equiv_class)




#!/usr/bin/env python

from setuptools import setup, find_packages

BASE_PKGS=find_packages('src', exclude=['drp', 'drp.*'])
NAMESPACE_PKGS = ['numina.pipelines', 'numina.pipelines.emir']
ALL_PKGS = BASE_PKGS + NAMESPACE_PKGS

setup(name='pyemir',
      version='0.6.4',
      author='Sergio Pascual',
      author_email='sergiopr@fis.ucm.es',
      url='http://guaix.fis.ucm.es/projects/emir',
      download_url='http://astrax.fis.ucm.es/software/pyemir/pyemir-0.6.4.tar.gz',
      license='GPLv3',
      description='EMIR Data Processing Pipeline',
      packages=ALL_PKGS,
      package_dir={'emir': 'src/emir', 'numina.pipelines': 'src/drp'},
      package_data={'emir.simulation': ['*.dat'],
                     'emir.instrument': ['image_*.txt', 'spectrum_*.txt'],
                   },
      test_suite="nose.collector",
      install_requires=['setuptools', 'numpy', 'pyfits', 
                        'scipy', 'sphinx', 'pywcs',
                        'matplotlib', 'numdisplay', 
                        'numina>=0.7.0'],
      test_requires=['nose',
                     ],
      # numdisplay lives here
      dependency_links = [
        'http://stsdas.stsci.edu/numdisplay'
        ],
      classifiers=[
                   "Programming Language :: Python :: 2.7",
                   'Development Status :: 3 - Alpha',
                   "Environment :: Other Environment",
                   "Intended Audience :: Science/Research",
                   "License :: OSI Approved :: GNU General Public License (GPL)",
                   "Operating System :: OS Independent",
                   "Topic :: Scientific/Engineering :: Astronomy",
                   ],
      long_description=open('README.txt').read()
      )



#
# Copyright 2008-2012 Universidad Complutense de Madrid
# 
# This file is part of PyEmir
# 
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

'''Recipe for the reduction of imaging mode observations.'''

import logging

from numina.recipes import RecipeBase, Parameter, DataProductParameter
from numina.recipes.requirements import Requirement
from numina.recipes import provides, DataFrame

from emir.dataproducts import MasterBias, MasterDark 
from emir.dataproducts import MasterIntensityFlat, MasterBadPixelMask
from emir.dataproducts import SourcesCatalog, NonLinearityCalibration

from .shared import DirectImageCommon


__author__ = "Sergio Pascual <sergiopr@fis.ucm.es>"

_logger = logging.getLogger("numina.recipes.emir")

@provides(DataFrame, SourcesCatalog)
class DitheredImageRecipe(RecipeBase, DirectImageCommon):
    '''Recipe for the reduction of imaging mode observations.

    Recipe to reduce observations obtained in imaging mode, considering different
    possibilities depending on the size of the offsets between individual images.
    In particular, the following observing modes are considered: stare imaging, nodded
    beamswitched imaging, and dithered imaging. 
    
    A critical piece of information here is a table that clearly specifies which
    images can be labeled as *science*, and which ones as *sky*. Note that some
    images are used both as *science* and *sky* (when the size of the targets are
    small compared to the offsets).
    
    **Observing modes:**
     
     * StareImage
     * Nodded/Beam-switched images
     * Dithered images 
    
    
    **Inputs:**
    
     * Science frames + [Sky Frames]
     * Observing mode name: **stare image**, **nodded beamswitched image**, or **dithered imaging**
     * A table relating each science image with its sky image(s) (TBD if it's in 
       the FITS header and/or in other format)
     * Offsets between them (Offsets must be integer)
     * Master Dark 
     * Bad pixel mask (BPM) 
     * Non-linearity correction polynomials 
     * Master flat (twilight/dome flats)
     * Master background (thermal background, only in K band)
     * Exposure Time (must be the same in all the frames)
     * Airmass for each frame
     * Detector model (gain, RN, lecture mode)
     * Average extinction in the filter
     * Astrometric calibration (TBD)
    
    **Outputs:**
    
     * Image with three extensions: final image scaled to the individual exposure
       time, variance  and exposure time map OR number of images combined (TBD)
    
    **Procedure:**
    
    Images are corrected from dark, non-linearity and flat. Then, an iterative
    process starts:
    
     * Sky is computed from each frame, using the list of sky images of each
       science frame. The objects are avoided using a mask (from the second
       iteration on).
    
     * The relative offsets are the nominal from the telescope. From the second
       iteration on, we refine them using objects of appropriate brightness (not
       too bright, not to faint).
    
     * We combine the sky-subtracted images, output is: a new image, a variance
       image and a exposure map/number of images used map.
    
     * An object mask is generated.
    
     * We recompute the sky map, using the object mask as an additional input. From
       here we iterate (typically 4 times).
    
     * Finally, the images are corrected from atmospheric extinction and flux
       calibrated.
    
     * A preliminary astrometric calibration can always be used (using the central
       coordinates of the pointing and the plate scale in the detector). A better
       calibration might be computed using available stars (TBD).
    
    '''

    __requires__ = [
        DataProductParameter('master_bpm', MasterBadPixelMask, 
                  'Master bad pixel mask'),       
        DataProductParameter('master_bias', MasterBias, 'Master bias image', optional=True),
        DataProductParameter('master_dark', MasterDark, 'Master dark image'),
        DataProductParameter('nonlinearity', NonLinearityCalibration([1.0, 0.0]), 
                  'Polynomial for non-linearity correction'),
        DataProductParameter('master_intensity_ff', MasterIntensityFlat, 
                  'Master intensity flatfield'),
        Parameter('extinction', 0.0, 'Mean atmospheric extinction'),
        # FIXME: this parameter is optional 
        Parameter('sources', None, 
                  'List of x, y coordinates to measure FWHM',
                  optional=True),
        Parameter('offsets', None, 'List of pairs of offsets',
                  optional=True),
        Parameter('iterations', 4, 'Iterations of the recipe'),
        Parameter('sky_images', 5, 'Images used to estimate the background before and after current image'),
        Parameter('sky_images_sep_time', 10, 'Maximum separation time between consecutive sky images in minutes'),
        Parameter('check_photometry_levels', [0.5, 0.8], 'Levels to check the flux of the objects'),
        Parameter('check_photometry_actions', ['warn', 'warn', 'default'], 'Actions to take on images'),
        Requirement('instrument.detector.channels', 'List of channels'),
        Requirement('instrument.detector.shape', 'Detector shape'),                    
                    
    ]

    def __init__(self):
        super(DitheredImageRecipe, self).__init__(
            author="Sergio Pascual <sergiopr@fis.ucm.es>",
            version="0.1.0"
        )
        
    def run(self, obresult):        
        baseshape = self.parameters['instrument.detector.shape']
        amplifiers = self.parameters['instrument.detector.channels']
        offsets = self.parameters['offsets']
        
        
        
        return self.process(obresult, baseshape, amplifiers, 
                            offsets=offsets, subpix=1, 
                            stop_after=DirectImageCommon.FULLRED)
                                                            



#
# Copyright 2011-2012 Universidad Complutense de Madrid
# 
# This file is part of PyEmir
# 
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

'''MOS Recipes for EMIR'''

import logging

from numina.recipes import RecipeBase, Parameter, provides
from numina.recipes import DataProductParameter

from ..dataproducts import MasterBias, MasterDark, MasterBadPixelMask 
from ..dataproducts import MasterIntensityFlat, MasterSpectralFlat
from ..dataproducts import (NonLinearityCalibration, SlitTransmissionCalibration,
WavelengthCalibration)
from ..dataproducts import Spectra, LinesCatalog, DataCube

__all__ = []

_logger = logging.getLogger('emir.recipes')

@provides(Spectra, LinesCatalog)
class StareSpectraRecipe(RecipeBase):
    '''Recipe for the reduction of multiobject spectroscopy.
    
    Recipe to reduce observations obtained in multiobject spectroscopy,
    considering different possibilities depending on the observing mode. In
    particular, the following observing modes are considered: stare spectra,
    dithered/nodded spectra, and offset spectra.
    
    The special flexibility of the Cold Slit Unit (CSU) allows to obtain slit
    configurations in which some of the individual slits can be perfectly aligned
    with some other slits. When these aligned slits are contiguous, they constitute
    what are going to be considered as **pseudo-long-slits**. In the particular
    case in which a given individual slit is not aligned with any other, the
    pseudo-long-slit will have the shortest length *Lslit*, the length of a single
    individual slit. In the rest of the cases, where the number of individual slits
    that are aligned consecutively is *Nslit*>1, the total length of the
    pseudo-long-slit will be *Nslit* x *Lslit*.
    
    In order to unambiguously proceed with a proper sky subtraction for different
    observing modes, a table specifying which pseudo-long-slits with *sky*
    spectra are associated to the pseudo-long-slits with *science* spectra must be
    provided. In addition, this table must specify the characteristics of the data
    obtained through each pseudo-long-slit. More precisely, in the case of
    *science* spectra, it must indicate whether the target is puntual or extended.
    The *sky* spectra must indicate the corresponding position and range (pixels
    along the slit, referred to the central pseudo-long-slit coordinates) of the
    valid sky regions. In this way, both *sky* and *science* spectra can be
    extracted from the same pseudo-long-slit, if this is required.
    
    The following possibilities must be considered at the time of creating the
    association table among pseudo-long-slits:
    
     * A single *sky* pseudo-long-slit can be associated to several *science*
       pseudo-long-slits.
     * The association between pseudo-long-slits is not restricted to
       pseudo-long-slits of the same spectroscopic image. For example, in the case
       of separate exposures for *science* and *sky* spectra, a given
       pseudo-long-slit can contain the *science* spectra in one exposure and the
       *sky* spectra in another exposure.
     * The configuration of the CSU must be identical when pseudo-long-slits of
       different images are associated.
    
    **Observing modes:**
    
     * Stare spectra
     * Dithered/Nodded spectra
     * Offset Spectra 
    
    **Inputs:**
    
     * Science frames + [Sky Frames]
     * An indication of the observing mode: **stare spectra**, 
       **dithered/nodded spectra**, or **offset spectra**. In the case of
       dithered/nodded spectra, the corresponding offsets must be indicated (these
       offsets must be integer and can be zero)
     * The table relating the associations among the different pseudo-long-slits
       and the *science* and *sky* intervals that will be extracted at the end of
       the data reduction (TBD if this table must be in the FITS header and/or in
       other format)
     * Master Dark 
     * Bad pixel mask (BPM) 
     * Non-linearity correction polynomials 
     * Spectroscopic Master flat (twilight/dome flats)
     * Master background (thermal background, only in K band)
     * Exposure Time (must be the same in all the frames, TBD)
     * Airmass for each frame
     * Detector model (gain, RN, lecture mode)
     * Average extinction curve the spectral range
     * Response curve for each pseudo-long-slit.
    
    **Outputs:**
    
     * 2D image with two extensions: final 2D image and associated 2D variance
       image, before sky-subtaction (TBD)
     * Final sky-subtracted and extracted spectra. TBD the way in which these
       spectra (and associated variance spectra) will be arranged (single 2D image,
       individual 1D spectra in different FITS extensions, individual 1D FITS files
       for each pseudo-long-slit)
    
    **Procedure:**
    
    The reduction will be carried out by extracting the different pseudo-long-slits
    and following a *traditional* long-slit reduction, making the appropriate
    treatment for the sky subtraction. The basic steps must include:
    
     * Data modeling (if appropriate/possible) and variance frame creation (from
       first principles)
    
     * Correction of non-linearity
    
     * Dark correction: this step can be avoided in most cases, since the
       extracted sky spectra may contain that information. However, if observing
       modes with different exposure times for *science* and *sky* frames are
       allowed, this correction may be necessary.
    
     * Flatfielding: distinguish between high frequency (pixel-to-pixel) and 
       low-frequency (overall response and slit illumination) corrections. 
       Lamp flats are adequate for the former and twilight flats for the second.
    
     * Detection and extraction of slits: apply Border_Detection algorithm, from
       own frames or from flatfields.
    
     * Cleaning
    
       * Single spectroscopic image: sigma-clipping algorithm removing 
         local background in pre-defined direction(s).
       * Multiple spectroscopic images: sigma-clipping from comparison between 
         frames.
     
     * Wavelength calibration and C-distortion correction of each slit. 
       Double-check with available sky lines.
    
     * Sky-subtraction (number of sources/slit will be allowed to be > 1?).
    
       * Subtraction using sky signal at the borders of the same slit.
       * Subtraction using sky signal from other(s) slit(s), not necessarily 
         adjacent.
     
     * Spectrophotometric calibration of each slit, using the extinction correction 
       curve and the master spectrophotometric calibration curve.
    
     * Spectra extraction: define optimal, average, peak, FWHM.
    
    '''


    __requires__ = [       
        DataProductParameter('master_bias', MasterBias, 'Master bias image'),
        DataProductParameter('master_dark', MasterDark, 'Master dark image'),
        DataProductParameter('master_bpm', MasterBadPixelMask, 'Master bad pixel mask'),
        DataProductParameter('nonlinearity', NonLinearityCalibration([1.0, 0.0]), 
                  'Polynomial for non-linearity correction'),
        DataProductParameter('master_intensity_ff', MasterIntensityFlat, 
                  'Master intensity flatfield'),
        DataProductParameter('master_spectral_ff', MasterSpectralFlat, 
                  'Master spectral flatfield'),
        DataProductParameter('st_calibration', SlitTransmissionCalibration, 
                  'Slit tranmision calibration'),
        DataProductParameter('w_calibration', WavelengthCalibration, 
                  'Wavelength calibration'),
        Parameter('lines', None, 
                  'List of x-lambda pairs of line coordinates'),        
    ]

    def __init__(self):
        super(StareSpectraRecipe, self).__init__(
            author="Sergio Pascual <sergiopr@fis.ucm.es>",
            version="0.1.0"
        )

    def run(self, obresult):
        return {'products': [Spectra(), LinesCatalog()]}

@provides(Spectra, LinesCatalog)
class DNSpectraRecipe(RecipeBase):
    '''
    Observing mode:
        Dithered/Nodded spectra along the slit
    '''
    
    __requires__ = [       
        DataProductParameter('master_bias', MasterBias, 'Master bias image'),
        DataProductParameter('master_dark', MasterDark, 'Master dark image'),
        DataProductParameter('master_bpm', MasterBadPixelMask, 'Master bad pixel mask'),
        DataProductParameter('nonlinearity', NonLinearityCalibration([1.0, 0.0]), 
                  'Polynomial for non-linearity correction'),
        DataProductParameter('master_intensity_ff', MasterIntensityFlat, 
                  'Master intensity flatfield'),
        DataProductParameter('master_spectral_ff', MasterSpectralFlat, 
                  'Master spectral flatfield'),
        DataProductParameter('st_calibration', SlitTransmissionCalibration, 
                  'Slit tranmision calibration'),
        DataProductParameter('w_calibration', WavelengthCalibration, 
                  'Wavelength calibration'),
        Parameter('lines', None, 
                  'List of x-lambda pairs of line coordinates'),        
    ]

    def __init__(self):
        super(DNSpectraRecipe, self).__init__(
            author="Sergio Pascual <sergiopr@fis.ucm.es>",
            version="0.1.0"
        )

    def run(self, obresult):
        return {'products': [Spectra(), LinesCatalog()]}
    
@provides(Spectra, LinesCatalog)
class OffsetSpectraRecipe(RecipeBase):
    '''
    Observing mode:
        Offset spectra beyond the slit
    '''
    
    __requires__ = [       
        DataProductParameter('master_bias', MasterBias, 'Master bias image'),
        DataProductParameter('master_dark', MasterDark, 'Master dark image'),
        DataProductParameter('master_bpm', MasterBadPixelMask, 'Master bad pixel mask'),
        DataProductParameter('nonlinearity', NonLinearityCalibration([1.0, 0.0]), 
                  'Polynomial for non-linearity correction'),
        DataProductParameter('master_intensity_ff', MasterIntensityFlat, 
                  'Master intensity flatfield'),
        DataProductParameter('master_spectral_ff', MasterSpectralFlat, 
                  'Master spectral flatfield'),
        DataProductParameter('st_calibration', SlitTransmissionCalibration, 
                  'Slit tranmision calibration'),
        DataProductParameter('w_calibration', WavelengthCalibration, 
                  'Wavelength calibration'),
        Parameter('lines', None, 
                  'List of x-lambda pairs of line coordinates'),        
    ]

    def __init__(self):
        super(OffsetSpectraRecipe, self).__init__(
            author="Sergio Pascual <sergiopr@fis.ucm.es>",
            version="0.1.0"
        )

    def run(self, obresult):
        return {'products': [Spectra(), LinesCatalog()]}
    
    
@provides(DataCube)
class RasterSpectraRecipe(RecipeBase):
    '''
    Observing mode:
        Raster spectra
    '''
    
    __requires__ = [       
        DataProductParameter('master_bias', MasterBias, 'Master bias image'),
        DataProductParameter('master_dark', MasterDark, 'Master dark image'),
        DataProductParameter('master_bpm', MasterBadPixelMask, 'Master bad pixel mask'),
        DataProductParameter('nonlinearity', NonLinearityCalibration([1.0, 0.0]), 
                  'Polynomial for non-linearity correction'),
        DataProductParameter('master_intensity_ff', MasterIntensityFlat, 
                  'Master intensity flatfield'),
        DataProductParameter('master_spectral_ff', MasterSpectralFlat, 
                  'Master spectral flatfield'),
        DataProductParameter('st_calibration', SlitTransmissionCalibration, 
                  'Slit tranmision calibration'),
        DataProductParameter('w_calibration', WavelengthCalibration, 
                  'Wavelength calibration')        
    ]

    def __init__(self):
        super(RasterSpectraRecipe, self).__init__(
            author="Sergio Pascual <sergiopr@fis.ucm.es>",
            version="0.1.0"
        )

    def run(self, obresult):
        return {'products': [DataCube()]}